#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Optional

from bravado.client import SwaggerClient
from bravado.exception import HTTPError

from neptune.new.exceptions import (
    NeptuneFieldCountLimitExceedException,
    NeptuneLimitExceedException,
)


class ApiMethodWrapper:
    ATTRIBUTES_PER_EXPERIMENT_LIMIT_EXCEEDED = "ATTRIBUTES_PER_EXPERIMENT_LIMIT_EXCEEDED"
    WORKSPACE_IN_READ_ONLY_MODE = "WORKSPACE_IN_READ_ONLY_MODE"
    PROJECT_KEY_COLLISION = "PROJECT_KEY_COLLISION"
    PROJECT_NAME_COLLISION = "PROJECT_NAME_COLLISION"
    PROJECT_KEY_INVALID = "PROJECT_KEY_INVALID"
    PROJECT_NAME_INVALID = "PROJECT_NAME_INVALID"

    def __init__(self, api_method):
        self._api_method = api_method

    @staticmethod
    def handle_neptune_http_errors(response, exception: Optional[HTTPError] = None):
        try:
            body = response.json() or dict()
        except Exception:
            body = {}

        error_type: Optional[str] = body.get("errorType")
        if error_type == ApiMethodWrapper.ATTRIBUTES_PER_EXPERIMENT_LIMIT_EXCEEDED:
            raise NeptuneFieldCountLimitExceedException(
                limit=body.get("limit", "<unknown limit>"),
                container_type=body.get("experimentType", "object"),
                identifier=body.get("experimentQualifiedName", "<unknown identifier>"),
            )
        elif error_type == ApiMethodWrapper.WORKSPACE_IN_READ_ONLY_MODE:
            raise NeptuneLimitExceedException(
                reason=body.get("title", "Unknown reason")
            ) from exception
        elif error_type == ApiMethodWrapper.PROJECT_KEY_COLLISION:
            from neptune.management.exceptions import ProjectKeyCollision

            raise ProjectKeyCollision(key=body.get("key", "<unknown key>")) from exception
        elif error_type == ApiMethodWrapper.PROJECT_NAME_COLLISION:
            from neptune.management.exceptions import ProjectNameCollision

            raise ProjectNameCollision(name=body.get("name", "<unknown name>")) from exception
        elif error_type == ApiMethodWrapper.PROJECT_KEY_INVALID:
            from neptune.management.exceptions import ProjectKeyInvalid

            raise ProjectKeyInvalid(
                key=body.get("key", "<unknown key>"), reason=body.get("reason", "Unknown reason")
            ) from exception
        elif error_type == ApiMethodWrapper.PROJECT_NAME_INVALID:
            from neptune.management.exceptions import ProjectNameInvalid

            raise ProjectNameInvalid(
                name=body.get("name", "<unknown name>"), reason=body.get("reason", "Unknown reason")
            ) from exception
        elif exception:
            raise exception

    def __call__(self, *args, **kwargs):
        try:
            future = self._api_method(*args, **kwargs)
            return FinishedApiResponseFuture(future.response())  # wait synchronously
        except HTTPError as e:
            self.handle_neptune_http_errors(e.response, exception=e)

    def __getattr__(self, item):
        return getattr(self._api_method, item)


class ApiWrapper:
    def __init__(self, api_obj):
        self._api_obj = api_obj

    def __getattr__(self, item):
        return ApiMethodWrapper(getattr(self._api_obj, item))


class FinishedApiResponseFuture:
    def __init__(self, response):
        self._response = response

    def response(self):
        return self._response


class SwaggerClientWrapper:
    def __init__(self, swagger_client: SwaggerClient):
        self._swagger_client = swagger_client
        self.api = ApiWrapper(swagger_client.api)
        self.swagger_spec = swagger_client.swagger_spec

    # For test purposes
    def __eq__(self, other):
        if isinstance(other, SwaggerClientWrapper):
            return self._swagger_client == other._swagger_client
        return False



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from neptune.backend import BackendApiClient
from neptune.exceptions import InvalidNeptuneBackend
from neptune.internal.api_clients import (
    HostedNeptuneBackendApiClient,
    OfflineBackendApiClient,
)


def backend_factory(*, backend_name, api_token=None, proxies=None) -> BackendApiClient:
    if backend_name == "offline":
        return OfflineBackendApiClient()

    elif backend_name is None:
        return HostedNeptuneBackendApiClient(api_token, proxies)

    else:
        raise InvalidNeptuneBackend(backend_name)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=redefined-outer-name
import os
import time

import boto3
import pytest
from faker import Faker

from e2e_tests.utils import (
    Environment,
    RawEnvironment,
    a_project_name,
    initialize_container,
)
from neptune.management import (
    add_project_member,
    add_project_service_account,
    create_project,
    delete_project,
)
from neptune.management.internal.utils import normalize_project_name

fake = Faker()


@pytest.fixture(scope="session")
def environment():
    raw_env = RawEnvironment()
    workspace = raw_env.workspace_name
    admin_token = raw_env.admin_neptune_api_token
    user = raw_env.user_username
    service_account_name = raw_env.service_account_name

    project_name = a_project_name(project_slug=fake.slug())
    project_identifier = normalize_project_name(name=project_name, workspace=workspace)
    created_project_identifier = create_project(
        name=project_name,
        visibility="priv",
        workspace=workspace,
        api_token=admin_token,
    )

    time.sleep(10)

    add_project_member(
        name=created_project_identifier,
        username=user,
        # pylint: disable=no-member
        role="contributor",
        api_token=admin_token,
    )

    add_project_service_account(
        name=created_project_identifier,
        service_account_name=service_account_name,
        # pylint: disable=no-member
        role="contributor",
        api_token=admin_token,
    )

    yield Environment(
        workspace=workspace,
        project=project_identifier,
        user_token=raw_env.neptune_api_token,
        admin_token=admin_token,
        admin=raw_env.admin_username,
        user=user,
        service_account=raw_env.service_account_name,
    )

    delete_project(name=created_project_identifier, api_token=admin_token)


@pytest.fixture(scope="session")
def container(request, environment):
    exp = initialize_container(container_type=request.param, project=environment.project)
    yield exp
    exp.stop()


@pytest.fixture(scope="session")
def containers_pair(request, environment):
    container_a_type, container_b_type = request.param.split("-")
    container_a = initialize_container(container_type=container_a_type, project=environment.project)
    container_b = initialize_container(container_type=container_b_type, project=environment.project)

    yield container_a, container_b

    container_b.stop()
    container_a.stop()


@pytest.fixture(scope="session")
def bucket(environment):
    bucket_name = os.environ.get("BUCKET_NAME")

    s3_client = boto3.resource("s3")
    s3_bucket = s3_client.Bucket(bucket_name)

    yield bucket_name, s3_client

    s3_bucket.objects.filter(Prefix=environment.project).delete()



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import threading
from pathlib import Path
from typing import Optional

from neptune.new.constants import ASYNC_DIRECTORY, OFFLINE_DIRECTORY
from neptune.new.exceptions import MetadataContainerNotFound
from neptune.new.internal.backends.api_model import ApiExperiment
from neptune.new.internal.container_type import ContainerType
from neptune.new.internal.disk_queue import DiskQueue
from neptune.new.internal.utils.sync_offset_file import SyncOffsetFile
from neptune.new.sync.utils import get_qualified_name
from tests.neptune.new.utils.api_experiments_factory import (
    api_metadata_container,
    api_run,
)


def generate_get_metadata_container(registered_containers):
    def get_metadata_container(container_id, expected_container_type: ContainerType):
        """This function will return run as well as projects. Will be cleaned in ModelRegistry"""
        for exp in registered_containers:
            if container_id in (str(exp.id), get_qualified_name(exp)):
                return exp

        raise MetadataContainerNotFound.of_container_type(
            container_type=expected_container_type, container_id=container_id
        )

    return get_metadata_container


# pylint: disable=unused-argument
def execute_operations(container_id, container_type, operations):
    return len(operations), []


def _prepare_disk_queue(*, exp_path, last_ack_version):
    exp_path.mkdir(parents=True)
    queue = DiskQueue(
        dir_path=exp_path,
        to_dict=lambda x: x,
        from_dict=lambda x: x,
        lock=threading.RLock(),
    )
    queue.put("op-0")
    queue.put("op-1")
    queue.put("op-2")

    SyncOffsetFile(exp_path / "last_put_version").write(3)
    if last_ack_version is not None:
        SyncOffsetFile(exp_path / "last_ack_version").write(last_ack_version)


def prepare_metadata_container(
    *, container_type: ContainerType, path: Path, last_ack_version: Optional[int]
) -> ApiExperiment:
    is_offline = last_ack_version is None

    container = api_metadata_container(container_type)

    if is_offline:
        exp_path = path / OFFLINE_DIRECTORY / f"{container.type.value}__{container.id}"
    else:
        execution_id = "exec-0"
        exp_path = path / ASYNC_DIRECTORY / f"{container.type.value}__{container.id}" / execution_id

    _prepare_disk_queue(
        exp_path=exp_path,
        last_ack_version=last_ack_version,
    )

    return container


def prepare_deprecated_run(*, path: Path, last_ack_version: Optional[int]):
    is_offline = last_ack_version is None

    run = api_run()

    if is_offline:
        exp_path = path / OFFLINE_DIRECTORY / run.id
    else:
        execution_id = "exec-0"
        exp_path = path / ASYNC_DIRECTORY / run.id / execution_id
    _prepare_disk_queue(
        exp_path=exp_path,
        last_ack_version=last_ack_version,
    )
    return run



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import re
from pathlib import Path

import pytest
from click.testing import CliRunner

import neptune.new as neptune
from e2e_tests.base import AVAILABLE_CONTAINERS, BaseE2ETest, fake
from e2e_tests.utils import (
    DISABLE_SYSLOG_KWARGS,
    initialize_container,
    reinitialize_container,
    tmp_context,
)
from neptune.new.exceptions import NeptuneException
from neptune.new.sync import sync

runner = CliRunner()


class TestSync(BaseE2ETest):
    SYNCHRONIZED_SYSID_RE = r"[\w-]+/[\w-]+/([\w-]+)"

    @pytest.mark.parametrize("container_type", AVAILABLE_CONTAINERS)
    def test_sync_container(self, container_type, environment):
        with tmp_context() as tmp:
            key = self.gen_key()
            original_value = fake.unique.word()
            updated_value = fake.unique.word()

            with initialize_container(
                container_type=container_type, project=environment.project
            ) as container:
                # assign original value
                container[key] = original_value
                container.wait()
                # pylint: disable=protected-access
                container_id = container._id
                container_sys_id = container._sys_id

            # manually add operations to queue
            queue_dir = list(
                Path(f"./.neptune/async/{container_type}__{container_id}/").glob("exec-*")
            )[0]
            with open(queue_dir / "last_put_version", encoding="utf-8") as last_put_version_f:
                last_put_version = int(last_put_version_f.read())
                with open(queue_dir / "data-1.log", "a", encoding="utf-8") as queue_f:
                    queue_f.write(
                        json.dumps(
                            {
                                "obj": {
                                    "type": "AssignString",
                                    "path": key.split("/"),
                                    "value": updated_value,
                                },
                                "version": last_put_version + 1,
                            }
                        )
                    )
                    queue_f.write(
                        json.dumps(
                            {
                                "obj": {
                                    "type": "CopyAttribute",
                                    "path": ["copy"] + key.split("/"),
                                    "container_id": container_id,
                                    "container_type": container_type,
                                    "source_path": key.split("/"),
                                    "source_attr_name": "String",
                                },
                                "version": last_put_version + 2,
                            }
                        )
                    )
                with open(
                    queue_dir / "last_put_version", "w", encoding="utf-8"
                ) as last_put_version_f:
                    last_put_version_f.write(str(last_put_version + 2))

            with reinitialize_container(
                container_sys_id, container_type, project=environment.project
            ) as container:
                # server should have the original value
                assert container[key].fetch() == original_value

            # run neptune sync
            result = runner.invoke(sync, ["--path", tmp])
            assert result.exit_code == 0

            with reinitialize_container(
                container_sys_id, container_type, project=environment.project
            ) as container:
                # and we should get the updated value from server
                assert container[key].fetch() == updated_value
                assert container["copy/" + key].fetch() == updated_value

    def test_offline_sync(self, environment):
        with tmp_context() as tmp:
            # create run in offline mode
            run = neptune.init(
                mode="offline",
                project=environment.project,
                **DISABLE_SYSLOG_KWARGS,
            )
            # assign some values
            key = self.gen_key()
            val = fake.word()
            run[key] = val

            # and stop it
            run.stop()

            # run asynchronously
            result = runner.invoke(sync, ["--path", tmp, "-p", environment.project])
            assert result.exit_code == 0

            # we'll have to parse sync output to determine short_id
            sys_id_found = re.search(self.SYNCHRONIZED_SYSID_RE, result.stdout)
            assert len(sys_id_found.groups()) == 1
            sys_id = sys_id_found.group(1)

            run2 = neptune.init_run(run=sys_id, project=environment.project)
            assert run2[key].fetch() == val

    @pytest.mark.parametrize("container_type", ["model", "model_version", "project"])
    def test_cannot_offline_non_runs(self, environment, container_type):
        with pytest.raises(NeptuneException) as e:
            initialize_container(
                container_type=container_type,
                project=environment.project,
                mode="offline",
            )
        assert "can't be initialized in OFFLINE mode" in str(e.value)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import abc
import atexit
import itertools
import threading
import time
import traceback
from contextlib import AbstractContextManager
from datetime import datetime
from functools import wraps
from typing import Any, Dict, List, Optional, Union

from neptune.exceptions import UNIX_STYLES
from neptune.new.attributes import create_attribute_from_type
from neptune.new.attributes.attribute import Attribute
from neptune.new.attributes.namespace import Namespace as NamespaceAttr
from neptune.new.attributes.namespace import NamespaceBuilder
from neptune.new.exceptions import (
    InactiveModelException,
    InactiveModelVersionException,
    InactiveProjectException,
    InactiveRunException,
    MetadataInconsistency,
    NeptunePossibleLegacyUsageException,
)
from neptune.new.handler import Handler
from neptune.new.internal.backends.api_model import AttributeType
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.backends.nql import NQLQuery
from neptune.new.internal.background_job import BackgroundJob
from neptune.new.internal.container_structure import ContainerStructure
from neptune.new.internal.container_type import ContainerType
from neptune.new.internal.id_formats import SysId, UniqueId
from neptune.new.internal.operation import DeleteAttribute
from neptune.new.internal.operation_processors.operation_processor import (
    OperationProcessor,
)
from neptune.new.internal.state import ContainerState
from neptune.new.internal.utils import (
    is_bool,
    is_dict_like,
    is_float,
    is_float_like,
    is_int,
    is_string,
    is_string_like,
    verify_type,
)
from neptune.new.internal.utils.logger import logger
from neptune.new.internal.utils.paths import parse_path
from neptune.new.internal.utils.runningmode import in_interactive, in_notebook
from neptune.new.internal.utils.uncaught_exception_handler import (
    instance as uncaught_exception_handler,
)
from neptune.new.internal.value_to_attribute_visitor import ValueToAttributeVisitor
from neptune.new.metadata_containers.metadata_containers_table import Table
from neptune.new.types import Boolean, Integer
from neptune.new.types.atoms.datetime import Datetime
from neptune.new.types.atoms.float import Float
from neptune.new.types.atoms.string import String
from neptune.new.types.mode import Mode
from neptune.new.types.namespace import Namespace
from neptune.new.types.value import Value
from neptune.new.types.value_copy import ValueCopy


def ensure_not_stopped(fun):
    @wraps(fun)
    def inner_fun(self: "MetadataContainer", *args, **kwargs):
        # pylint: disable=protected-access
        if self._state == ContainerState.STOPPED:
            if self.container_type == ContainerType.RUN:
                raise InactiveRunException(label=self._label)
            elif self.container_type == ContainerType.PROJECT:
                raise InactiveProjectException(label=self._label)
            elif self.container_type == ContainerType.MODEL:
                raise InactiveModelException(label=self._label)
            elif self.container_type == ContainerType.MODEL_VERSION:
                raise InactiveModelVersionException(label=self._label)
            else:
                raise ValueError(f"Unknown container type: {self.container_type}")
        return fun(self, *args, **kwargs)

    return inner_fun


class MetadataContainer(AbstractContextManager):
    container_type: ContainerType

    LEGACY_METHODS = set()

    def __init__(
        self,
        *,
        id_: UniqueId,
        mode: Mode,
        backend: NeptuneBackend,
        op_processor: OperationProcessor,
        background_job: BackgroundJob,
        lock: threading.RLock,
        project_id: UniqueId,
        project_name: str,
        workspace: str,
        sys_id: SysId,
    ):
        self._id = id_
        self._mode = mode
        self._project_id = project_id
        self._project_name = project_name
        self._workspace = workspace
        self._backend = backend
        self._op_processor = op_processor
        self._bg_job = background_job
        self._structure: ContainerStructure[Attribute, NamespaceAttr] = ContainerStructure(
            NamespaceBuilder(self)
        )
        self._lock = lock
        self._state = ContainerState.CREATED
        self._sys_id = sys_id

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_tb is not None:
            traceback.print_exception(exc_type, exc_val, exc_tb)
        self.stop()

    def __getattr__(self, item):
        if item in self.LEGACY_METHODS:
            raise NeptunePossibleLegacyUsageException()
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}'")

    @property
    @abc.abstractmethod
    def _label(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _docs_url_stop(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _url(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _metadata_url(self) -> str:
        raise NotImplementedError

    def _get_subpath_suggestions(self, path_prefix: str = None, limit: int = 1000) -> List[str]:
        parsed_path = parse_path(path_prefix or "")
        return list(itertools.islice(self._structure.iterate_subpaths(parsed_path), limit))

    def _ipython_key_completions_(self):
        return self._get_subpath_suggestions()

    @ensure_not_stopped
    def __getitem__(self, path: str) -> "Handler":
        return Handler(self, path)

    @ensure_not_stopped
    def __setitem__(self, key: str, value) -> None:
        self.__getitem__(key).assign(value)

    @ensure_not_stopped
    def __delitem__(self, path) -> None:
        self.pop(path)

    @ensure_not_stopped
    def assign(self, value, wait: bool = False) -> None:
        self._get_root_handler().assign(value, wait)

    @ensure_not_stopped
    def fetch(self) -> dict:
        return self._get_root_handler().fetch()

    def ping(self):
        self._backend.ping(self._id, self.container_type)

    def start(self):
        atexit.register(self._shutdown_hook)
        self._op_processor.start()
        self._bg_job.start(self)
        self._state = ContainerState.STARTED

    def stop(self, seconds: Optional[Union[float, int]] = None) -> None:
        verify_type("seconds", seconds, (float, int, type(None)))
        if self._state != ContainerState.STARTED:
            return

        self._state = ContainerState.STOPPING
        ts = time.time()
        logger.info("Shutting down background jobs, please wait a moment...")
        self._bg_job.stop()
        self._bg_job.join(seconds)
        logger.info("Done!")
        with self._lock:
            sec_left = None if seconds is None else seconds - (time.time() - ts)
            self._op_processor.stop(sec_left)
        if self._mode != Mode.OFFLINE:
            logger.info("Explore the metadata in the Neptune app:")
            logger.info(self._metadata_url)
        self._backend.close()
        self._state = ContainerState.STOPPED

    def get_structure(self) -> Dict[str, Any]:
        # This is very weird pylint false-positive.
        # pylint: disable=no-member
        return self._structure.get_structure().to_dict()

    def print_structure(self) -> None:
        self._print_structure_impl(self.get_structure(), indent=0)

    def _print_structure_impl(self, struct: dict, indent: int) -> None:
        for key in sorted(struct.keys()):
            print("    " * indent, end="")
            if isinstance(struct[key], dict):
                print(
                    "{blue}'{key}'{end}:".format(
                        blue=UNIX_STYLES["blue"], key=key, end=UNIX_STYLES["end"]
                    )
                )
                self._print_structure_impl(struct[key], indent=indent + 1)
            else:
                print(
                    "{blue}'{key}'{end}: {type}".format(
                        blue=UNIX_STYLES["blue"],
                        key=key,
                        end=UNIX_STYLES["end"],
                        type=type(struct[key]).__name__,
                    )
                )

    def define(
        self,
        path: str,
        value: Union[Value, int, float, str, datetime],
        wait: bool = False,
    ) -> Attribute:
        if isinstance(value, Value):
            pass
        elif isinstance(value, Handler):
            value = ValueCopy(value)
        elif is_bool(value):
            value = Boolean(value)
        elif is_int(value):
            value = Integer(value)
        elif is_float(value):
            value = Float(value)
        elif is_string(value):
            value = String(value)
        elif isinstance(value, datetime):
            value = Datetime(value)
        elif is_float_like(value):
            value = Float(float(value))
        elif is_dict_like(value):
            value = Namespace(value)
        elif is_string_like(value):
            value = String(str(value))
        else:
            raise TypeError("Value of unsupported type {}".format(type(value)))
        parsed_path = parse_path(path)

        with self._lock:
            old_attr = self._structure.get(parsed_path)
            if old_attr:
                raise MetadataInconsistency(
                    "Attribute or namespace {} is already defined".format(path)
                )
            attr = ValueToAttributeVisitor(self, parsed_path).visit(value)
            self._structure.set(parsed_path, attr)
            attr.process_assignment(value, wait)
            return attr

    def get_attribute(self, path: str) -> Optional[Attribute]:
        with self._lock:
            return self._structure.get(parse_path(path))

    def set_attribute(self, path: str, attribute: Attribute) -> Optional[Attribute]:
        with self._lock:
            return self._structure.set(parse_path(path), attribute)

    def exists(self, path: str) -> bool:
        verify_type("path", path, str)
        return self.get_attribute(path) is not None

    @ensure_not_stopped
    def pop(self, path: str, wait: bool = False) -> None:
        verify_type("path", path, str)
        self._get_root_handler().pop(path, wait)

    def _pop_impl(self, parsed_path: List[str], wait: bool):
        self._structure.pop(parsed_path)
        self._op_processor.enqueue_operation(DeleteAttribute(parsed_path), wait)

    def lock(self) -> threading.RLock:
        return self._lock

    def wait(self, disk_only=False) -> None:
        with self._lock:
            if disk_only:
                self._op_processor.flush()
            else:
                self._op_processor.wait()

    def sync(self, wait: bool = True) -> None:
        with self._lock:
            if wait:
                self._op_processor.wait()
            attributes = self._backend.get_attributes(self._id, self.container_type)
            self._structure.clear()
            for attribute in attributes:
                self._define_attribute(parse_path(attribute.path), attribute.type)

    def _define_attribute(self, _path: List[str], _type: AttributeType):
        attr = create_attribute_from_type(_type, self, _path)
        self._structure.set(_path, attr)

    def _get_root_handler(self):
        return Handler(self, "")

    def get_url(self) -> str:
        """Returns the URL that can be accessed within the browser"""
        return self._url

    def _startup(self, debug_mode):
        if not debug_mode:
            logger.info(self.get_url())

        self.start()

        if not debug_mode:
            if in_interactive() or in_notebook():
                logger.info(
                    "Remember to stop your %s once you’ve finished logging your metadata (%s)."
                    " It will be stopped automatically only when the notebook"
                    " kernel/interactive console is terminated.",
                    self.container_type.value,
                    self._docs_url_stop,
                )

        uncaught_exception_handler.activate()

    def _shutdown_hook(self):
        self.stop()

    def _fetch_entries(self, child_type: ContainerType, query: NQLQuery) -> Table:
        leaderboard_entries = self._backend.search_leaderboard_entries(
            project_id=self._project_id,
            types=[child_type],
            query=query,
        )

        return Table(
            backend=self._backend,
            container_type=child_type,
            entries=leaderboard_entries,
        )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["StatusRunner"]

import sys
import textwrap
from pathlib import Path
from typing import List, Sequence, Tuple

from neptune.new.constants import ASYNC_DIRECTORY, OFFLINE_NAME_PREFIX
from neptune.new.envs import PROJECT_ENV_NAME
from neptune.new.internal.backends.api_model import ApiExperiment
from neptune.new.internal.utils.logger import logger
from neptune.new.sync.abstract_backend_runner import AbstractBackendRunner
from neptune.new.sync.utils import (
    get_metadata_container,
    get_offline_dirs,
    get_qualified_name,
    is_container_synced,
    iterate_containers,
)

offline_run_explainer = """
Runs which execute offline are not created on the server and they are not assigned to projects;
instead, they are identified by UUIDs like the ones above.
When synchronizing offline runs, please specify the workspace and project using the "--project"
flag. Alternatively, you can set the environment variable
{} to the target workspace/project. See the examples below.
""".format(
    PROJECT_ENV_NAME
)


class StatusRunner(AbstractBackendRunner):
    def partition_containers(
        self,
        base_path: Path,
    ) -> Tuple[List[ApiExperiment], List[ApiExperiment], int]:
        synced_containers = []
        unsynced_containers = []
        async_path = base_path / ASYNC_DIRECTORY
        for container_type, container_id, path in iterate_containers(async_path):
            metadata_container = get_metadata_container(
                backend=self._backend,
                container_id=container_id,
                container_type=container_type,
            )

            if is_container_synced(path):
                synced_containers.append(metadata_container)
            else:
                unsynced_containers.append(metadata_container)

        not_found = len(
            [exp for exp in synced_containers + unsynced_containers if not exp or exp.trashed]
        )
        synced_containers = [obj for obj in synced_containers if obj and not obj.trashed]
        unsynced_containers = [obj for obj in unsynced_containers if obj and not obj.trashed]

        return synced_containers, unsynced_containers, not_found

    @staticmethod
    def list_containers(
        base_path: Path,
        synced_containers: Sequence[ApiExperiment],
        unsynced_containers: Sequence[ApiExperiment],
        offline_dirs: Sequence[str],
    ) -> None:
        if not synced_containers and not unsynced_containers and not offline_dirs:
            logger.info("There are no Neptune objects in %s", base_path)
            sys.exit(1)

        if unsynced_containers:
            logger.info("Unsynchronized objects:")
            for container in unsynced_containers:
                logger.info("- %s", get_qualified_name(container))

        if synced_containers:
            logger.info("Synchronized objects:")
            for container in synced_containers:
                logger.info("- %s", get_qualified_name(container))

        if offline_dirs:
            logger.info("Unsynchronized offline objects:")
            for run_id in offline_dirs:
                logger.info("- %s", f"{OFFLINE_NAME_PREFIX}{run_id}")
            logger.info("\n%s", textwrap.fill(offline_run_explainer, width=90))

        if not unsynced_containers:
            logger.info("\nThere are no unsynchronized objects in %s", base_path)

        if not synced_containers:
            logger.info("\nThere are no synchronized objects in %s", base_path)

        logger.info("\nPlease run with the `neptune sync --help` to see example commands.")

    def synchronization_status(self, base_path: Path) -> None:
        synced_containers, unsynced_containers, not_found = self.partition_containers(base_path)
        if not_found > 0:
            logger.warning(
                "WARNING: %s objects was skipped because they are in trash or do not exist anymore.",
                not_found,
            )
        offline_dirs = get_offline_dirs(base_path)
        self.list_containers(base_path, synced_containers, unsynced_containers, offline_dirs)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Dict

import pytest

from e2e_tests.base import BaseE2ETest, fake
from e2e_tests.utils import Environment, a_project_name
from neptune.management import (
    add_project_member,
    add_project_service_account,
    create_project,
    delete_project,
    get_project_list,
    get_project_member_list,
    get_project_service_account_list,
    get_workspace_member_list,
    get_workspace_service_account_list,
    remove_project_member,
    remove_project_service_account,
)
from neptune.management.exceptions import UserNotExistsOrWithoutAccess
from neptune.management.internal.utils import normalize_project_name


@pytest.mark.management
class TestManagement(BaseE2ETest):
    @staticmethod
    def _assure_presence_and_role(
        *, username: str, expected_role: str, member_list: Dict[str, str]
    ):
        assert username in member_list
        assert member_list.get(username) == expected_role

    def test_standard_scenario(self, environment: Environment):
        project_name = a_project_name(project_slug=f"{fake.slug()}-mgmt")
        project_identifier = normalize_project_name(
            name=project_name, workspace=environment.workspace
        )

        assert project_identifier not in get_project_list(api_token=environment.admin_token)
        assert project_identifier not in get_project_list(api_token=environment.user_token)

        self._assure_presence_and_role(
            username=environment.user,
            expected_role="member",
            member_list=get_workspace_member_list(
                name=environment.workspace, api_token=environment.admin_token
            ),
        )
        self._assure_presence_and_role(
            username=environment.user,
            expected_role="member",
            member_list=get_workspace_member_list(
                name=environment.workspace, api_token=environment.user_token
            ),
        )
        self._assure_presence_and_role(
            username=environment.service_account,
            expected_role="member",
            member_list=get_workspace_service_account_list(
                name=environment.workspace, api_token=environment.user_token
            ),
        )

        created_project_identifier = create_project(
            name=project_name,
            visibility="priv",
            workspace=environment.workspace,
            api_token=environment.admin_token,
        )

        assert created_project_identifier == project_identifier
        assert created_project_identifier in get_project_list(api_token=environment.admin_token)
        assert created_project_identifier not in get_project_list(api_token=environment.user_token)

        assert environment.user not in get_project_member_list(
            name=created_project_identifier, api_token=environment.admin_token
        )
        assert environment.service_account not in get_project_service_account_list(
            name=created_project_identifier, api_token=environment.admin_token
        )

        add_project_service_account(
            name=created_project_identifier,
            service_account_name=environment.service_account,
            role="contributor",
            api_token=environment.admin_token,
        )
        add_project_member(
            name=created_project_identifier,
            username=environment.user,
            role="contributor",
            api_token=environment.admin_token,
        )

        project_members = get_project_member_list(
            name=created_project_identifier, api_token=environment.admin_token
        )
        assert environment.user in project_members
        assert project_members.get(environment.user) == "contributor"

        project_members = get_project_member_list(
            name=created_project_identifier, api_token=environment.user_token
        )
        assert environment.user in project_members
        assert project_members.get(environment.user) == "contributor"
        assert environment.service_account not in project_members

        assert created_project_identifier in get_project_list(api_token=environment.user_token)

        remove_project_member(
            name=created_project_identifier,
            username=environment.user,
            api_token=environment.admin_token,
        )
        remove_project_service_account(
            name=created_project_identifier,
            service_account_name=environment.service_account,
            api_token=environment.admin_token,
        )

        assert created_project_identifier not in get_project_list(api_token=environment.user_token)
        assert environment.user not in get_project_member_list(
            name=created_project_identifier, api_token=environment.admin_token
        )
        assert environment.service_account not in get_project_service_account_list(
            name=created_project_identifier, api_token=environment.admin_token
        )

        delete_project(name=created_project_identifier, api_token=environment.admin_token)

        assert created_project_identifier not in get_project_list(api_token=environment.admin_token)

    def test_visibility_workspace(self, environment: "Environment"):
        project_name = a_project_name(project_slug=f"{fake.slug()}-workspace")
        project_identifier = normalize_project_name(
            name=project_name, workspace=environment.workspace
        )

        assert project_identifier not in get_project_list(api_token=environment.admin_token)
        assert project_identifier not in get_project_list(api_token=environment.user_token)

        self._assure_presence_and_role(
            username=environment.user,
            expected_role="member",
            member_list=get_workspace_member_list(
                name=environment.workspace, api_token=environment.admin_token
            ),
        )

        created_project_identifier = create_project(
            name=project_name,
            visibility="workspace",
            workspace=environment.workspace,
            api_token=environment.admin_token,
        )

        assert created_project_identifier == project_identifier
        assert created_project_identifier in get_project_list(api_token=environment.admin_token)

        self._assure_presence_and_role(
            username=environment.user,
            expected_role="owner",
            member_list=get_project_member_list(
                name=created_project_identifier, api_token=environment.admin_token
            ),
        )
        assert environment.service_account not in get_project_service_account_list(
            name=created_project_identifier, api_token=environment.admin_token
        )

        add_project_service_account(
            name=created_project_identifier,
            service_account_name=environment.service_account,
            role="contributor",
            api_token=environment.admin_token,
        )

        self._assure_presence_and_role(
            username=environment.service_account,
            expected_role="contributor",
            member_list=get_project_service_account_list(
                name=created_project_identifier, api_token=environment.admin_token
            ),
        )

        with pytest.raises(UserNotExistsOrWithoutAccess):
            remove_project_member(
                name=created_project_identifier,
                username=environment.user,
                api_token=environment.admin_token,
            )

        remove_project_service_account(
            name=created_project_identifier,
            service_account_name=environment.service_account,
            api_token=environment.admin_token,
        )

        self._assure_presence_and_role(
            username=environment.user,
            expected_role="owner",
            member_list=get_project_member_list(
                name=created_project_identifier, api_token=environment.admin_token
            ),
        )
        assert environment.service_account not in get_project_service_account_list(
            name=created_project_identifier, api_token=environment.admin_token
        )

        delete_project(name=created_project_identifier, api_token=environment.admin_token)

        assert project_identifier not in get_project_list(api_token=environment.admin_token)

    def test_create_project(self, environment: "Environment"):
        project_name = a_project_name(project_slug=f"{fake.slug()}-create")
        project_identifier = normalize_project_name(
            name=project_name, workspace=environment.workspace
        )

        assert project_identifier not in get_project_list(api_token=environment.user_token)
        self._assure_presence_and_role(
            username=environment.user,
            expected_role="member",
            member_list=get_workspace_member_list(
                name=environment.workspace, api_token=environment.user_token
            ),
        )

        created_project_identifier = create_project(
            name=project_name,
            workspace=environment.workspace,
            api_token=environment.user_token,
        )

        assert created_project_identifier == project_identifier
        assert created_project_identifier in get_project_list(api_token=environment.user_token)

        delete_project(name=created_project_identifier, api_token=environment.admin_token)

        assert project_identifier not in get_project_list(api_token=environment.user_token)

    def _test_add_sa_to_project_as_owner(
        self, created_project_identifier: str, environment: "Environment"
    ):
        self._assure_presence_and_role(
            username=environment.user,
            expected_role="owner",
            member_list=get_project_member_list(
                name=created_project_identifier, api_token=environment.user_token
            ),
        )

        assert environment.service_account not in get_project_service_account_list(
            name=created_project_identifier, api_token=environment.user_token
        )

        add_project_service_account(
            name=created_project_identifier,
            service_account_name=environment.service_account,
            role="contributor",
            api_token=environment.user_token,
        )
        self._assure_presence_and_role(
            username=environment.service_account,
            expected_role="contributor",
            member_list=get_project_service_account_list(
                name=created_project_identifier, api_token=environment.user_token
            ),
        )

        remove_project_service_account(
            name=created_project_identifier,
            service_account_name=environment.service_account,
            api_token=environment.user_token,
        )
        assert environment.service_account not in get_project_service_account_list(
            name=created_project_identifier, api_token=environment.admin_token
        )

    def _test_add_user_to_project_as_sa(
        self, created_project_identifier: str, environment: "Environment"
    ):
        self._assure_presence_and_role(
            username=environment.service_account,
            expected_role="owner",
            member_list=get_project_service_account_list(
                name=created_project_identifier, api_token=environment.user_token
            ),
        )

        assert environment.user not in get_project_member_list(
            name=created_project_identifier, api_token=environment.user_token
        )

        add_project_member(
            name=created_project_identifier,
            username=environment.user,
            role="contributor",
            api_token=environment.admin_token,
        )
        self._assure_presence_and_role(
            username=environment.user,
            expected_role="contributor",
            member_list=get_project_member_list(
                name=created_project_identifier, api_token=environment.user_token
            ),
        )

        remove_project_member(
            name=created_project_identifier,
            username=environment.user,
            api_token=environment.admin_token,
        )
        assert environment.user not in get_project_member_list(
            name=created_project_identifier, api_token=environment.user_token
        )

    def test_invite_as_non_admin(self, environment: "Environment"):
        project_name = a_project_name(project_slug=f"{fake.slug()}-invitation")
        project_identifier = normalize_project_name(
            name=project_name, workspace=environment.workspace
        )

        created_project_identifier = create_project(
            name=project_name,
            workspace=environment.workspace,
            api_token=environment.user_token,
        )

        assert created_project_identifier == project_identifier
        assert created_project_identifier in get_project_list(api_token=environment.user_token)

        # user who created a project (`user_token` owner) will be automatically project owner
        sa_is_project_owner = (
            get_project_service_account_list(
                name=created_project_identifier, api_token=environment.user_token
            ).get(environment.service_account)
            == "owner"
        )
        user_is_project_owner = (
            get_project_member_list(
                name=created_project_identifier, api_token=environment.user_token
            ).get(environment.user)
            == "owner"
        )
        if sa_is_project_owner and not user_is_project_owner:
            # SA has access to project, so tests are run as SA
            self._test_add_user_to_project_as_sa(created_project_identifier, environment)
        elif user_is_project_owner and not sa_is_project_owner:
            # SA doesn't have access to project, so tests are run as user
            self._test_add_sa_to_project_as_owner(created_project_identifier, environment)
        else:
            raise AssertionError(
                "Expected to only SA or user to be owner of newly created project."
            )

        delete_project(name=created_project_identifier, api_token=environment.admin_token)

        assert project_identifier not in get_project_list(api_token=environment.user_token)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from neptune.new import init_model
from neptune.new.internal.container_type import ContainerType
from tests.neptune.new.client.abstract_tables_test import AbstractTablesTestMixin


class TestModelVersionTables(AbstractTablesTestMixin, unittest.TestCase):
    expected_container_type = ContainerType.MODEL_VERSION

    def get_table(self):
        return init_model(
            model="organization/project",
            project="PRO-MOD",
            mode="read-only",
        ).fetch_model_versions_table()

    def get_table_entries(self, table):
        return table.to_rows()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import socket
import unittest
import uuid
from unittest.mock import call

from bravado.exception import HTTPNotFound, HTTPPaymentRequired, HTTPUnprocessableEntity
from mock import MagicMock, patch
from packaging.version import Version

from neptune.new.exceptions import (
    CannotResolveHostname,
    FileUploadError,
    MetadataInconsistency,
    NeptuneClientUpgradeRequiredError,
    NeptuneLimitExceedException,
)
from neptune.new.internal.backends.hosted_client import (  # pylint:disable=protected-access
    DEFAULT_REQUEST_KWARGS,
    _get_token_client,
    create_artifacts_client,
    create_backend_client,
    create_http_client_with_auth,
    create_leaderboard_client,
    get_client_config,
)
from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.new.internal.backends.swagger_client_wrapper import SwaggerClientWrapper
from neptune.new.internal.backends.utils import verify_host_resolution
from neptune.new.internal.container_type import ContainerType
from neptune.new.internal.credentials import Credentials
from neptune.new.internal.operation import (
    AssignString,
    LogFloats,
    TrackFilesToArtifact,
    UploadFile,
    UploadFileContent,
)
from neptune.new.internal.utils import base64_encode
from tests.neptune.new.backend_test_mixin import BackendTestMixin
from tests.neptune.new.utils import response_mock

API_TOKEN = (
    "eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vYXBwLnN0YWdlLm5lcHR1bmUuYWkiLCJ"
    "hcGlfa2V5IjoiOTJhNzhiOWQtZTc3Ni00ODlhLWI5YzEtNzRkYmI1ZGVkMzAyIn0="
)

credentials = Credentials.from_token(API_TOKEN)


@patch("neptune.new.internal.backends.hosted_client.RequestsClient", new=MagicMock())
@patch("neptune.new.internal.backends.hosted_client.NeptuneAuthenticator", new=MagicMock())
@patch("bravado.client.SwaggerClient.from_url")
@patch("platform.platform", new=lambda: "testPlatform")
@patch("platform.python_version", new=lambda: "3.9.test")
class TestHostedNeptuneBackend(unittest.TestCase, BackendTestMixin):
    # pylint:disable=protected-access

    def setUp(self) -> None:
        # Clear all LRU storage
        verify_host_resolution.cache_clear()
        _get_token_client.cache_clear()
        get_client_config.cache_clear()
        create_http_client_with_auth.cache_clear()
        create_backend_client.cache_clear()
        create_leaderboard_client.cache_clear()
        create_artifacts_client.cache_clear()

        self.container_types = [ContainerType.RUN, ContainerType.PROJECT]

    @patch("neptune.new.internal.backends.hosted_neptune_backend.upload_file_attribute")
    @patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_execute_operations(self, upload_mock, swagger_client_factory):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        container_uuid = str(uuid.uuid4())

        response_error = MagicMock()
        response_error.errorDescription = "error1"
        swagger_client.api.executeOperations().response().result = [response_error]
        swagger_client.api.executeOperations.reset_mock()
        upload_mock.return_value = [FileUploadError("file1", "error2")]
        some_text = "Some streamed text"
        some_binary = b"Some streamed binary"

        for container_type in self.container_types:
            with self.subTest(msg=f"For type {container_type.value}"):
                upload_mock.reset_mock()
                swagger_client_factory.reset_mock()

                # when
                result = backend.execute_operations(
                    container_id=container_uuid,
                    container_type=container_type,
                    operations=[
                        UploadFile(
                            path=["some", "files", "some_file"],
                            ext="",
                            file_path="path_to_file",
                        ),
                        UploadFileContent(
                            path=["some", "files", "some_text_stream"],
                            ext="txt",
                            file_content=base64_encode(some_text.encode("utf-8")),
                        ),
                        UploadFileContent(
                            path=["some", "files", "some_binary_stream"],
                            ext="bin",
                            file_content=base64_encode(some_binary),
                        ),
                        LogFloats(["images", "img1"], [LogFloats.ValueType(1, 2, 3)]),
                        AssignString(["properties", "name"], "some text"),
                        UploadFile(
                            path=["some", "other", "file.txt"],
                            ext="txt",
                            file_path="other/file/path.txt",
                        ),
                    ],
                )

                # then
                swagger_client.api.executeOperations.assert_called_once_with(
                    **{
                        "experimentId": str(container_uuid),
                        "operations": [
                            {
                                "path": "images/img1",
                                "logFloats": {
                                    "entries": [
                                        {
                                            "value": 1,
                                            "step": 2,
                                            "timestampMilliseconds": 3000,
                                        }
                                    ]
                                },
                            },
                            {
                                "path": "properties/name",
                                "assignString": {"value": "some text"},
                            },
                        ],
                        **DEFAULT_REQUEST_KWARGS,
                    }
                )

                upload_mock.assert_has_calls(
                    [
                        call(
                            swagger_client=backend.leaderboard_client,
                            container_id=container_uuid,
                            attribute="some/other/file.txt",
                            source="other/file/path.txt",
                            ext="txt",
                            multipart_config=backend._client_config.multipart_config,
                        ),
                        call(
                            swagger_client=backend.leaderboard_client,
                            container_id=container_uuid,
                            attribute="some/files/some_file",
                            source="path_to_file",
                            ext="",
                            multipart_config=backend._client_config.multipart_config,
                        ),
                        call(
                            swagger_client=backend.leaderboard_client,
                            container_id=container_uuid,
                            attribute="some/files/some_text_stream",
                            source=some_text.encode("utf-8"),
                            ext="txt",
                            multipart_config=backend._client_config.multipart_config,
                        ),
                        call(
                            swagger_client=backend.leaderboard_client,
                            container_id=container_uuid,
                            attribute="some/files/some_binary_stream",
                            source=some_binary,
                            ext="bin",
                            multipart_config=backend._client_config.multipart_config,
                        ),
                    ],
                    any_order=True,
                )

                self.assertEqual(
                    (
                        6,
                        [
                            FileUploadError("file1", "error2"),
                            FileUploadError("file1", "error2"),
                            FileUploadError("file1", "error2"),
                            FileUploadError("file1", "error2"),
                            MetadataInconsistency("error1"),
                        ],
                    ),
                    result,
                )

    @patch("neptune.new.internal.backends.hosted_neptune_backend.upload_file_attribute")
    @patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_upload_files_destination_path(self, upload_mock, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        container_uuid = str(uuid.uuid4())

        for container_type in self.container_types:
            with self.subTest(msg=f"For type {container_type.value}"):
                upload_mock.reset_mock()
                swagger_client_factory.reset_mock()

                # when
                backend.execute_operations(
                    container_id=container_uuid,
                    container_type=container_type,
                    operations=[
                        UploadFile(
                            path=["some", "path", "1", "var"],
                            ext="",
                            file_path="/path/to/file",
                        ),
                        UploadFile(
                            path=["some", "path", "2", "var"],
                            ext="txt",
                            file_path="/some.file/with.dots.txt",
                        ),
                        UploadFile(
                            path=["some", "path", "3", "var"],
                            ext="jpeg",
                            file_path="/path/to/some_image.jpeg",
                        ),
                    ],
                )

                # then
                upload_mock.assert_has_calls(
                    [
                        call(
                            swagger_client=backend.leaderboard_client,
                            container_id=container_uuid,
                            attribute="some/path/1/var",
                            source="/path/to/file",
                            ext="",
                            multipart_config=backend._client_config.multipart_config,
                        ),
                        call(
                            swagger_client=backend.leaderboard_client,
                            container_id=container_uuid,
                            attribute="some/path/2/var",
                            source="/some.file/with.dots.txt",
                            ext="txt",
                            multipart_config=backend._client_config.multipart_config,
                        ),
                        call(
                            swagger_client=backend.leaderboard_client,
                            container_id=container_uuid,
                            attribute="some/path/3/var",
                            source="/path/to/some_image.jpeg",
                            ext="jpeg",
                            multipart_config=backend._client_config.multipart_config,
                        ),
                    ],
                    any_order=True,
                )

    @patch("neptune.new.internal.backends.hosted_neptune_backend.track_to_new_artifact")
    @patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_track_to_new_artifact(self, track_to_new_artifact_mock, swagger_client_factory):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        container_id = str(uuid.uuid4())
        project_id = str(uuid.uuid4())

        response_error = MagicMock()
        response_error.errorDescription = "error1"
        swagger_client.api.executeOperations.return_value.response.return_value.result = [
            response_error
        ]
        swagger_client.api.getArtifactAttribute.side_effect = HTTPNotFound(response=response_mock())
        swagger_client_wrapper = SwaggerClientWrapper(swagger_client)

        for container_type in self.container_types:
            with self.subTest(msg=f"For type {container_type.value}"):
                track_to_new_artifact_mock.reset_mock()
                swagger_client_factory.reset_mock()

                # when
                backend.execute_operations(
                    container_id=container_id,
                    container_type=container_type,
                    operations=[
                        TrackFilesToArtifact(
                            path=["sub", "one"],
                            project_id=project_id,
                            entries=[("/path/to/file", "/path/to")],
                        ),
                        TrackFilesToArtifact(
                            path=["sub", "two"],
                            project_id=project_id,
                            entries=[
                                ("/path/to/file1", None),
                                ("/path/to/file2", None),
                            ],
                        ),
                        TrackFilesToArtifact(
                            path=["sub", "three"],
                            project_id=project_id,
                            entries=[("/path/to/file1", None)],
                        ),
                        TrackFilesToArtifact(
                            path=["sub", "three"],
                            project_id=project_id,
                            entries=[("/path/to/file2", None)],
                        ),
                    ],
                )

                # then
                track_to_new_artifact_mock.assert_has_calls(
                    [
                        call(
                            swagger_client=swagger_client_wrapper,
                            project_id=project_id,
                            path=["sub", "one"],
                            parent_identifier=str(container_id),
                            entries=[("/path/to/file", "/path/to")],
                            default_request_params=DEFAULT_REQUEST_KWARGS,
                        ),
                        call(
                            swagger_client=swagger_client_wrapper,
                            project_id=project_id,
                            path=["sub", "two"],
                            parent_identifier=str(container_id),
                            entries=[
                                ("/path/to/file1", None),
                                ("/path/to/file2", None),
                            ],
                            default_request_params=DEFAULT_REQUEST_KWARGS,
                        ),
                        call(
                            swagger_client=swagger_client_wrapper,
                            project_id=project_id,
                            path=["sub", "three"],
                            parent_identifier=str(container_id),
                            entries=[
                                ("/path/to/file1", None),
                                ("/path/to/file2", None),
                            ],
                            default_request_params=DEFAULT_REQUEST_KWARGS,
                        ),
                    ],
                    any_order=True,
                )

    @patch("neptune.new.internal.backends.hosted_neptune_backend.track_to_existing_artifact")
    @patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_track_to_existing_artifact(
        self, track_to_existing_artifact_mock, swagger_client_factory
    ):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        container_id = str(uuid.uuid4())
        project_id = str(uuid.uuid4())

        response_error = MagicMock()
        response_error.errorDescription = "error1"
        swagger_client.api.executeOperations.return_value.response.return_value.result = [
            response_error
        ]
        swagger_client.api.getArtifactAttribute.return_value.response.return_value.result.hash = (
            "dummyHash"
        )
        swagger_client_wrapper = SwaggerClientWrapper(swagger_client)

        for container_type in self.container_types:
            track_to_existing_artifact_mock.reset_mock()
            swagger_client_factory.reset_mock()

            with self.subTest(msg=f"For type {container_type.value}"):
                track_to_existing_artifact_mock.reset_mock()
                swagger_client_factory.reset_mock()

                # when
                backend.execute_operations(
                    container_id=container_id,
                    container_type=container_type,
                    operations=[
                        TrackFilesToArtifact(
                            path=["sub", "one"],
                            project_id=project_id,
                            entries=[("/path/to/file", "/path/to")],
                        ),
                        TrackFilesToArtifact(
                            path=["sub", "two"],
                            project_id=project_id,
                            entries=[
                                ("/path/to/file1", None),
                                ("/path/to/file2", None),
                            ],
                        ),
                        TrackFilesToArtifact(
                            path=["sub", "three"],
                            project_id=project_id,
                            entries=[("/path/to/file1", None)],
                        ),
                        TrackFilesToArtifact(
                            path=["sub", "three"],
                            project_id=project_id,
                            entries=[("/path/to/file2", None)],
                        ),
                    ],
                )

                # then
                track_to_existing_artifact_mock.assert_has_calls(
                    [
                        call(
                            swagger_client=swagger_client_wrapper,
                            project_id=project_id,
                            path=["sub", "one"],
                            artifact_hash="dummyHash",
                            parent_identifier=str(container_id),
                            entries=[("/path/to/file", "/path/to")],
                            default_request_params=DEFAULT_REQUEST_KWARGS,
                        ),
                        call(
                            swagger_client=swagger_client_wrapper,
                            project_id=project_id,
                            path=["sub", "two"],
                            artifact_hash="dummyHash",
                            parent_identifier=str(container_id),
                            entries=[
                                ("/path/to/file1", None),
                                ("/path/to/file2", None),
                            ],
                            default_request_params=DEFAULT_REQUEST_KWARGS,
                        ),
                        call(
                            swagger_client=swagger_client_wrapper,
                            project_id=project_id,
                            path=["sub", "three"],
                            artifact_hash="dummyHash",
                            parent_identifier=str(container_id),
                            entries=[
                                ("/path/to/file1", None),
                                ("/path/to/file2", None),
                            ],
                            default_request_params=DEFAULT_REQUEST_KWARGS,
                        ),
                    ],
                    any_order=True,
                )

    @patch(
        "neptune.new.internal.backends.hosted_client.neptune_client_version",
        Version("0.5.13"),
    )
    @patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_min_compatible_version_ok(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, min_compatible="0.5.13")

        # expect
        HostedNeptuneBackend(credentials)

    @patch(
        "neptune.new.internal.backends.hosted_client.neptune_client_version",
        Version("0.5.13"),
    )
    @patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_min_compatible_version_fail(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, min_compatible="0.5.14")

        # expect
        with self.assertRaises(NeptuneClientUpgradeRequiredError) as ex:
            HostedNeptuneBackend(credentials)

        self.assertTrue("minimum required version is >=0.5.14" in str(ex.exception))

    @patch(
        "neptune.new.internal.backends.hosted_client.neptune_client_version",
        Version("0.5.13"),
    )
    @patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_max_compatible_version_ok(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, max_compatible="0.5.12")

        # expect
        HostedNeptuneBackend(credentials)

    @patch(
        "neptune.new.internal.backends.hosted_client.neptune_client_version",
        Version("0.5.13"),
    )
    @patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_max_compatible_version_fail(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, max_compatible="0.4.999")

        # expect
        with self.assertRaises(NeptuneClientUpgradeRequiredError) as ex:
            HostedNeptuneBackend(credentials)

        self.assertTrue("minimum required version is ==0.4.0" in str(ex.exception))

    @patch("socket.gethostbyname")
    def test_cannot_resolve_host(self, gethostname_mock, _):
        # given
        gethostname_mock.side_effect = socket.gaierror

        # expect
        with self.assertRaises(CannotResolveHostname):
            HostedNeptuneBackend(credentials)

    @patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_limit_exceed(self, swagger_client_factory):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        container_uuid = str(uuid.uuid4())

        # when:
        error = response_mock()
        error.json.return_value = {"title": "Maximum storage limit reached"}
        swagger_client.api.executeOperations.side_effect = HTTPPaymentRequired(response=error)

        # then:
        for container_type in self.container_types:
            with self.subTest(msg=f"For type {container_type.value}"):
                with self.assertRaises(NeptuneLimitExceedException):
                    backend.execute_operations(
                        container_id=container_uuid,
                        container_type=container_type,
                        operations=[
                            LogFloats(["float1"], [LogFloats.ValueType(1, 2, 3)]),
                        ],
                    )

    @patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_limit_exceed_legacy(self, swagger_client_factory):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        container_uuid = str(uuid.uuid4())

        # when:
        error = response_mock()
        error.json.return_value = {"title": "Monitoring hours not left"}
        swagger_client.api.executeOperations.side_effect = HTTPUnprocessableEntity(response=error)

        # then:
        for container_type in self.container_types:
            with self.subTest(msg=f"For type {container_type.value}"):
                with self.assertRaises(NeptuneLimitExceedException):
                    backend.execute_operations(
                        container_id=container_uuid,
                        container_type=container_type,
                        operations=[
                            LogFloats(["float1"], [LogFloats.ValueType(1, 2, 3)]),
                        ],
                    )



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import socket
import unittest

import mock
from mock import MagicMock

from neptune.exceptions import (
    CannotResolveHostname,
    DeprecatedApiToken,
    UnsupportedClientVersion,
)
from neptune.internal.api_clients import HostedNeptuneBackendApiClient

API_TOKEN = (
    "eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vYXBwLnN0YWdlLm5lcHR1bmUuYWkiLCJ"
    "hcGlfa2V5IjoiOTJhNzhiOWQtZTc3Ni00ODlhLWI5YzEtNzRkYmI1ZGVkMzAyIn0="
)


@mock.patch(
    "neptune.internal.api_clients.hosted_api_clients.hosted_backend_api_client.NeptuneAuthenticator",
    new=MagicMock,
)
class TestHostedNeptuneBackend(unittest.TestCase):
    # pylint:disable=protected-access

    @mock.patch("bravado.client.SwaggerClient.from_url")
    @mock.patch("neptune.__version__", "0.5.13")
    @mock.patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_min_compatible_version_ok(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, min_compatible="0.5.13")

        # expect
        HostedNeptuneBackendApiClient(api_token=API_TOKEN)

    @mock.patch("bravado.client.SwaggerClient.from_url")
    @mock.patch("neptune.__version__", "0.5.13")
    @mock.patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_min_compatible_version_fail(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, min_compatible="0.5.14")

        # expect
        with self.assertRaises(UnsupportedClientVersion) as ex:
            HostedNeptuneBackendApiClient(api_token=API_TOKEN)

        self.assertTrue("Please install neptune-client>=0.5.14" in str(ex.exception))

    @mock.patch("bravado.client.SwaggerClient.from_url")
    @mock.patch("neptune.__version__", "0.5.13")
    @mock.patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_max_compatible_version_ok(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, max_compatible="0.5.13")

        # expect
        HostedNeptuneBackendApiClient(api_token=API_TOKEN)

    @mock.patch("bravado.client.SwaggerClient.from_url")
    @mock.patch("neptune.__version__", "0.5.13")
    @mock.patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_max_compatible_version_fail(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, max_compatible="0.5.12")

        # expect
        with self.assertRaises(UnsupportedClientVersion) as ex:
            HostedNeptuneBackendApiClient(api_token=API_TOKEN)

        self.assertTrue("Please install neptune-client==0.5.12" in str(ex.exception))

    # pylint: disable=unused-argument
    @mock.patch("bravado.client.SwaggerClient.from_url")
    @mock.patch("neptune.internal.api_clients.credentials.os.getenv", return_value=API_TOKEN)
    @mock.patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_should_take_default_credentials_from_env(self, env, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory)

        # when
        backend = HostedNeptuneBackendApiClient()

        # then
        self.assertEqual(API_TOKEN, backend.credentials.api_token)

    @mock.patch("bravado.client.SwaggerClient.from_url")
    @mock.patch("socket.gethostbyname", MagicMock(return_value="1.1.1.1"))
    def test_should_accept_given_api_token(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory)

        # when
        session = HostedNeptuneBackendApiClient(API_TOKEN)

        # then
        self.assertEqual(API_TOKEN, session.credentials.api_token)

    @mock.patch("socket.gethostbyname")
    def test_deprecated_token(self, gethostname_mock):
        # given
        token = (
            "eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vdWkuc3RhZ2UubmVwdHVuZS5tbCIsImF"
            "waV9rZXkiOiI5ODM4ZDk1NC00MDAzLTExZTktYmY1MC0yMzE5ODM1NWRhNjYifQ=="
        )

        gethostname_mock.side_effect = socket.gaierror

        # expect
        with self.assertRaises(DeprecatedApiToken):
            HostedNeptuneBackendApiClient(token)

    @mock.patch("socket.gethostbyname")
    def test_cannot_resolve_host(self, gethostname_mock):
        # given
        token = (
            "eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vdWkuc3RhZ2UubmVwdHVuZS5tbCIsImFwaV91cmwiOiJodHRwczovL3VpLn"
            "N0YWdlLm5lcHR1bmUuYWkiLCJhcGlfa2V5IjoiOTgzOGQ5NTQtNDAwMy0xMWU5LWJmNTAtMjMxOTgzNTVkYTY2In0="
        )

        gethostname_mock.side_effect = socket.gaierror

        # expect
        with self.assertRaises(CannotResolveHostname):
            HostedNeptuneBackendApiClient(token)

    @staticmethod
    def _get_swagger_client_mock(
        swagger_client_factory,
        min_recommended=None,
        min_compatible=None,
        max_compatible=None,
    ):
        py_lib_versions = type("py_lib_versions", (object,), {})()
        setattr(py_lib_versions, "minRecommendedVersion", min_recommended)
        setattr(py_lib_versions, "minCompatibleVersion", min_compatible)
        setattr(py_lib_versions, "maxCompatibleVersion", max_compatible)

        artifacts = type("artifacts", (object,), {})()
        setattr(artifacts, "enabled", True)

        multipart_upload = type("multiPartUpload", (object,), {})()
        setattr(multipart_upload, "enabled", True)
        setattr(multipart_upload, "minChunkSize", 5242880)
        setattr(multipart_upload, "maxChunkSize", 1073741824)
        setattr(multipart_upload, "maxChunkCount", 1000)
        setattr(multipart_upload, "maxSinglePartSize", 5242880)

        client_config = type("client_config_response_result", (object,), {})()
        setattr(client_config, "pyLibVersions", py_lib_versions)
        setattr(client_config, "artifacts", artifacts)
        setattr(client_config, "multiPartUpload", multipart_upload)
        setattr(client_config, "apiUrl", None)
        setattr(client_config, "applicationUrl", None)

        swagger_client = MagicMock()
        swagger_client.api.getClientConfig.return_value.response.return_value.result = client_config
        swagger_client_factory.return_value = swagger_client

        return swagger_client


class SomeClass(object):
    pass


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re

import pytest
import pytorch_lightning as pl

from e2e_tests.base import BaseE2ETest


@pytest.mark.integrations
class TestPytorchLightning(BaseE2ETest):
    def test_logging_values(self, pytorch_run):
        # correct integration version is logged
        if pytorch_run.exists("source_code/integrations/lightning"):
            logged_version = pytorch_run["source_code/integrations/lightning"].fetch()
        else:
            logged_version = pytorch_run["source_code/integrations/pytorch-lightning"].fetch()
        assert logged_version == pl.__version__  # pylint: disable=E1101

        assert pytorch_run.exists("custom_prefix/valid/loss")
        assert len(pytorch_run["custom_prefix/valid/loss"].fetch_values()) == 3

    def test_saving_models(self, pytorch_run):
        best_model_path = pytorch_run["custom_prefix/model/best_model_path"].fetch()
        assert re.match(
            r".*my_model/checkpoints/epoch=.*-valid/loss=.*\.ckpt$",
            best_model_path,
        )
        assert pytorch_run["custom_prefix/model/best_model_score"].fetch() is not None

        # make sure that exactly `save_top_k` checkpoints
        # NOTE: when `max_epochs` is close to `save_top_k` there may be less than `save_top_k` saved models
        checkpoints = pytorch_run["custom_prefix/model/checkpoints"].fetch()
        assert all((checkpoint.startswith("epoch=") for checkpoint in checkpoints))
        assert len(checkpoints) == 2



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import random
import uuid
from itertools import product
from typing import Set
from unittest.mock import Mock
from zipfile import ZipFile

import pytest

from e2e_tests.base import AVAILABLE_CONTAINERS, BaseE2ETest, fake
from e2e_tests.utils import initialize_container, tmp_context
from neptune.new.internal.backends import hosted_file_operations
from neptune.new.internal.backends.api_model import MultipartConfig, OptionalFeatures
from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.new.metadata_containers import MetadataContainer
from neptune.new.types import File, FileSet


class TestUpload(BaseE2ETest):
    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_using_new_api(self, container: MetadataContainer):
        # pylint: disable=protected-access
        assert isinstance(container._backend, HostedNeptuneBackend)
        assert container._backend._client_config.has_feature(OptionalFeatures.MULTIPART_UPLOAD)
        assert isinstance(container._backend._client_config.multipart_config, MultipartConfig)

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    @pytest.mark.parametrize(
        "file_size",
        [
            pytest.param(10 * 2**20, id="big"),  # 10 MB, multipart
            pytest.param(100 * 2**10, id="small"),  # 100 kB, single upload
        ],
    )
    def test_single_file(self, container: MetadataContainer, file_size: int):
        key = self.gen_key()
        extension = fake.file_extension()
        filename = fake.file_name(extension=extension)
        downloaded_filename = fake.file_name()

        with tmp_context():
            # create file_size file
            with open(filename, "wb") as file:
                file.write(b"\0" * file_size)
            container[key].upload(filename)

            container.sync()
            container[key].download(downloaded_filename)

            assert container[key].fetch_extension() == extension
            assert os.path.getsize(downloaded_filename) == file_size
            with open(downloaded_filename, "rb") as file:
                content = file.read()
                assert len(content) == file_size
                assert content == b"\0" * file_size

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    @pytest.mark.parametrize(
        "file_size",
        [
            pytest.param(10 * 2**20, id="big"),  # 10 MB, multipart
            pytest.param(100 * 2**10, id="small"),  # 100 kB, single upload
        ],
    )
    def test_in_memory_file(self, container: MetadataContainer, file_size: int):
        key = self.gen_key()
        extension = fake.file_extension()
        downloaded_filename = fake.file_name()
        expected_content = os.urandom(file_size)

        container[key].upload(File.from_content(expected_content, extension))

        container.sync()
        container[key].download(downloaded_filename)

        assert container[key].fetch_extension() == extension
        assert os.path.getsize(downloaded_filename) == file_size
        with open(downloaded_filename, "rb") as file:
            content = file.read()
            assert len(content) == file_size
            assert content == expected_content

    def test_single_file_changed_during_upload(self, environment, monkeypatch):
        key = self.gen_key()
        file_size = 11 * 2**20  # 11 MB, multipart with 3 parts
        intermediate_size = 6 * 2**20  # 6 MB, second part < 5MB
        filename = fake.file_name()
        downloaded_filename = fake.file_name()

        _upload_raw_data = hosted_file_operations.upload_raw_data

        run = initialize_container(
            container_type="run",
            project=environment.project,
            mode="sync",
        )

        with tmp_context():
            # create file_size file
            with open(filename, "wb") as file:
                file.write(b"\0" * file_size)

            class UploadedFileChanger:
                def __init__(self):
                    self.upload_part_iteration = 0

                def __call__(self, *args, **kwargs):
                    # file starts to change and after uploading first part it's at intermediate_size
                    if self.upload_part_iteration == 0:
                        with open(filename, "wb") as file:
                            file.write(b"\0" * intermediate_size)
                    # after that it's back at file_size
                    elif self.upload_part_iteration == 1:
                        with open(filename, "wb") as file:
                            file.write(b"\0" * file_size)
                    self.upload_part_iteration += 1

                    return _upload_raw_data(*args, **kwargs)

            hacked_upload_raw_data = UploadedFileChanger()
            monkeypatch.setattr(
                hosted_file_operations,
                "upload_raw_data",
                Mock(wraps=hacked_upload_raw_data),
            )

            run[key].upload(filename)

            run[key].download(downloaded_filename)

            assert os.path.getsize(downloaded_filename) == file_size
            with open(downloaded_filename, "rb") as file:
                content = file.read()
                assert len(content) == file_size
                assert content == b"\0" * file_size
            # handling restart + 3 for actual upload
            assert hacked_upload_raw_data.upload_part_iteration == 5

    @pytest.mark.parametrize("container", ["run"], indirect=True)
    def test_replace_float_attribute_with_uploaded_file(self, container: MetadataContainer):
        key = self.gen_key()
        file_size = 100 * 2**10  # 100 kB
        filename = fake.file_name()
        downloaded_filename = fake.file_name()

        with tmp_context():
            # create file_size file
            with open(filename, "wb") as file:
                file.write(b"\0" * file_size)

            # set key to a float and sync it separately
            container[key] = 42.0
            container.sync()

            # delete and upload in the same queue flush
            container[key].pop()
            container[key].upload(filename)

            container.sync()
            container[key].download(downloaded_filename)

            assert os.path.getsize(downloaded_filename) == file_size
            with open(downloaded_filename, "rb") as file:
                content = file.read()
                assert len(content) == file_size
                assert content == b"\0" * file_size

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_fileset(self, container: MetadataContainer):
        key = self.gen_key()
        large_filesize = 10 * 2**20  # 10MB
        large_filename = fake.file_name()
        small_files = [
            (f"{uuid.uuid4()}.{fake.file_extension()}", fake.sentence().encode("utf-8"))
            for _ in range(100)
        ]

        with tmp_context():
            # create single large file (multipart) and a lot of very small files
            with open(large_filename, "wb") as file:
                file.write(b"\0" * large_filesize)
            for filename, contents in small_files:
                with open(filename, "wb") as file:
                    file.write(contents)

            small_filenames = [filename for filename, _ in small_files]
            # make sure there are no duplicates
            assert len({large_filename, *small_filenames}) == len(small_files) + 1

            # when one file as fileset uploaded
            container[key].upload_files([large_filename])

            # then check if will be downloaded
            container.sync()
            container[key].download("downloaded1.zip")

            with ZipFile("downloaded1.zip") as zipped:
                assert set(zipped.namelist()) == {large_filename, "/"}
                with zipped.open(large_filename, "r") as file:
                    content = file.read()
                    assert len(content) == large_filesize
                    assert content == b"\0" * large_filesize

            # when small files as fileset uploaded
            container[key].upload_files(small_filenames)

            # then check if everything will be downloaded
            container.sync()
            container[key].download("downloaded2.zip")

            with ZipFile("downloaded2.zip") as zipped:
                assert set(zipped.namelist()) == {large_filename, "/", *small_filenames}
                with zipped.open(large_filename, "r") as file:
                    content = file.read()
                    assert len(content) == large_filesize
                    assert content == b"\0" * large_filesize
                for filename, expected_content in small_files:
                    with zipped.open(filename, "r") as file:
                        content = file.read()
                        assert len(content) == len(expected_content)
                        assert content == expected_content

            # when first file is removed
            container[key].delete_files([large_filename])

            # then check if the rest will be downloaded
            container.sync()
            container[key].download("downloaded3.zip")

            with ZipFile("downloaded3.zip") as zipped:
                assert set(zipped.namelist()) == {"/", *small_filenames}
                for filename, expected_content in small_files:
                    with zipped.open(filename, "r") as file:
                        content = file.read()
                        assert len(content) == len(expected_content)
                        assert content == expected_content

    @classmethod
    def _gen_tree_paths(cls, depth, width=3) -> Set:
        """Generates all subdirectories of some random tree directory structure"""
        this_level_dirs = (fake.word() + "/" for _ in range(width))
        if depth == 1:
            return set(this_level_dirs)
        else:
            subpaths = cls._gen_tree_paths(depth=depth - 1, width=width)
            new_paths = set("".join(prod) for prod in product(subpaths, this_level_dirs))
            subpaths.update(new_paths)
            return subpaths

    @pytest.mark.parametrize("container", ["project", "run"], indirect=True)
    def test_fileset_nested_structure(self, container: MetadataContainer):
        key = self.gen_key()
        possible_paths = self._gen_tree_paths(depth=3)

        small_files = [
            (
                f"{path}{uuid.uuid4()}.{fake.file_extension()}",
                os.urandom(random.randint(10**3, 10**6)),
            )
            for path in possible_paths
        ]

        with tmp_context():
            # create dirs
            for dir_path in possible_paths:
                os.makedirs(dir_path, exist_ok=True)
            # create  a lot of very small files in different directories
            for filename, contents in small_files:
                with open(filename, "wb") as file:
                    file.write(contents)

            small_filenames = [filename for filename, _ in small_files]
            # make sure there are no duplicates
            assert len({*small_filenames}) == len(small_files)

            # when small files as fileset uploaded
            container[key].upload_files(".")

            # then check if everything will be downloaded
            container.sync()
            container[key].download("downloaded.zip")

            with ZipFile("downloaded.zip") as zipped:
                assert set(zipped.namelist()) == {
                    "/",
                    *possible_paths,
                    *small_filenames,
                }
                for filename, expected_content in small_files:
                    with zipped.open(filename, "r") as file:
                        content = file.read()
                        assert len(content) == len(expected_content)
                        assert content == expected_content

    @pytest.mark.parametrize("container", ["project", "run"], indirect=True)
    def test_reset_fileset(self, container: MetadataContainer):
        key = self.gen_key()
        filename1 = fake.file_name()
        filename2 = fake.file_name()
        content1 = os.urandom(random.randint(10**3, 10**6))
        content2 = os.urandom(random.randint(10**3, 10**6))

        with tmp_context():
            # create file1 and file2
            with open(filename1, "wb") as file1, open(filename2, "wb") as file2:
                file1.write(content1)
                file2.write(content2)

            # upload file1 to initial fileset
            container[key].upload_files(filename1)

            # then replace [file1] set with [file2] to the same key
            container.sync()
            container[key] = FileSet([filename2])

            # check if there's content of SECOND uploaded file
            container.sync()
            container[key].download("downloaded.zip")
            with ZipFile("downloaded.zip") as zipped:
                assert set(zipped.namelist()) == {filename2, "/"}
                with zipped.open(filename2, "r") as file:
                    content = file.read()
                    assert len(content) == len(content2)
                    assert content == content2

    @pytest.mark.parametrize("container", ["project", "run"], indirect=True)
    @pytest.mark.parametrize("delete_attribute", [True, False])
    def test_single_file_override(self, container: MetadataContainer, delete_attribute: bool):
        key = self.gen_key()
        filename1 = fake.file_name()
        filename2 = fake.file_name()
        content1 = os.urandom(random.randint(10**3, 10**6))
        content2 = os.urandom(random.randint(10**3, 10**6))
        downloaded_filename = fake.file_name()

        with tmp_context():
            # create file1 and file2
            with open(filename1, "wb") as file1, open(filename2, "wb") as file2:
                file1.write(content1)
                file2.write(content2)

            # upload file1 to key
            container[key].upload(filename1)

            if delete_attribute:
                # delete attribute
                del container[key]
                # make sure that attribute does not exist
                container.sync()
                with pytest.raises(AttributeError):
                    container[key].download(downloaded_filename)

            # then upload file2 to the same key
            container[key].upload(filename2)

            # check if there's content of SECOND uploaded file
            container.sync()
            container[key].download(downloaded_filename)
            with open(downloaded_filename, "rb") as file:
                content = file.read()
                assert len(content) == len(content2)
                assert content == content2

    @pytest.mark.parametrize("container", ["project", "run"], indirect=True)
    @pytest.mark.parametrize("delete_attribute", [True, False])
    def test_fileset_file_override(self, container: MetadataContainer, delete_attribute: bool):
        key = self.gen_key()
        filename = fake.file_name()
        content1 = os.urandom(random.randint(10**3, 10**6))
        content2 = os.urandom(random.randint(10**3, 10**6))

        with tmp_context():
            # create file
            with open(filename, "wb") as file1:
                file1.write(content1)
            # upload file1 to key
            container[key].upload_files([filename])

            if delete_attribute:
                # delete attribute
                del container[key]
                # make sure that attribute does not exist
                container.sync()
                with pytest.raises(AttributeError):
                    container[key].download("failed_download.zip")

            # override file content
            with open(filename, "wb") as file:
                file.write(content2)
            # then upload file2 to the same key
            container[key].upload_files([filename])

            # check if there's content of ONLY SECOND uploaded file
            container.sync()
            container[key].download("downloaded.zip")

            with ZipFile("downloaded.zip") as zipped:
                assert set(zipped.namelist()) == {filename, "/"}
                with zipped.open(filename, "r") as file:
                    content = file.read()
                    assert len(content) == len(content2)
                    assert content == content2



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

from neptune.new.internal.backends.nql import (
    NQLAggregator,
    NQLAttributeOperator,
    NQLAttributeType,
    NQLQueryAggregate,
    NQLQueryAttribute,
)


class TestNQL(unittest.TestCase):
    def test_attributes(self):
        self.assertEqual(
            str(
                NQLQueryAttribute(
                    name="sys/owner",
                    type=NQLAttributeType.STRING,
                    operator=NQLAttributeOperator.EQUALS,
                    value="user1",
                )
            ),
            '(`sys/owner`:string = "user1")',
        )
        self.assertEqual(
            str(
                NQLQueryAttribute(
                    name="sys/tags",
                    type=NQLAttributeType.STRING_SET,
                    operator=NQLAttributeOperator.CONTAINS,
                    value="tag1",
                )
            ),
            '(`sys/tags`:stringSet CONTAINS "tag1")',
        )
        self.assertEqual(
            str(
                NQLQueryAttribute(
                    name="sys/state",
                    type=NQLAttributeType.EXPERIMENT_STATE,
                    operator=NQLAttributeOperator.EQUALS,
                    value="running",
                )
            ),
            '(`sys/state`:experimentState = "running")',
        )

    def test_multiple_attribute_values(self):
        self.assertEqual(
            str(
                NQLQueryAggregate(
                    items=[
                        NQLQueryAttribute(
                            name="sys/owner",
                            type=NQLAttributeType.STRING,
                            operator=NQLAttributeOperator.EQUALS,
                            value=user,
                        )
                        for user in ["user1", "user2"]
                    ],
                    aggregator=NQLAggregator.OR,
                )
            ),
            '((`sys/owner`:string = "user1") OR (`sys/owner`:string = "user2"))',
        )

    def test_multiple_queries(self):
        self.assertEqual(
            str(
                NQLQueryAggregate(
                    items=[
                        NQLQueryAggregate(
                            items=[
                                NQLQueryAttribute(
                                    name="sys/owner",
                                    type=NQLAttributeType.STRING,
                                    operator=NQLAttributeOperator.EQUALS,
                                    value=user,
                                )
                                for user in ["user1", "user2"]
                            ],
                            aggregator=NQLAggregator.OR,
                        ),
                        NQLQueryAggregate(
                            items=[
                                NQLQueryAttribute(
                                    name="sys/tags",
                                    type=NQLAttributeType.STRING_SET,
                                    operator=NQLAttributeOperator.CONTAINS,
                                    value=tag,
                                )
                                for tag in ["tag1", "tag2"]
                            ],
                            aggregator=NQLAggregator.OR,
                        ),
                    ],
                    aggregator=NQLAggregator.AND,
                )
            ),
            '(((`sys/owner`:string = "user1") OR (`sys/owner`:string = "user2")) AND '
            '((`sys/tags`:stringSet CONTAINS "tag1") OR (`sys/tags`:stringSet CONTAINS "tag2")))',
        )



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import random

import pytest
from PIL import Image

from e2e_tests.base import AVAILABLE_CONTAINERS, BaseE2ETest, fake
from e2e_tests.utils import generate_image, image_to_png, tmp_context
from neptune.new.metadata_containers import MetadataContainer


class TestSeries(BaseE2ETest):
    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_log_numbers(self, container: MetadataContainer):
        key = self.gen_key()
        values = [random.random() for _ in range(50)]

        container[key].log(values[0])
        container[key].log(values[1:])
        container.sync()

        assert container[key].fetch_last() == values[-1]

        fetched_values = container[key].fetch_values()
        assert list(fetched_values["value"]) == values

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_log_strings(self, container: MetadataContainer):
        key = self.gen_key()
        values = [fake.word() for _ in range(50)]

        container[key].log(values[0])
        container[key].log(values[1:])
        container.sync()

        assert container[key].fetch_last() == values[-1]

        fetched_values = container[key].fetch_values()
        assert list(fetched_values["value"]) == values

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_log_images(self, container: MetadataContainer):
        key = self.gen_key()
        # images with size between 200KB - 12MB
        images = list(generate_image(size=2**n) for n in range(8, 12))

        container[key].log(images[0])
        container[key].log(images[1:])
        container.sync()

        with tmp_context():
            container[key].download_last("last")
            container[key].download("all")

            with Image.open("last/3.png") as img:
                assert img == image_to_png(image=images[-1])

            for i in range(4):
                with Image.open(f"all/{i}.png") as img:
                    assert img == image_to_png(image=images[i])



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest
import torch
from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.loggers.neptune import NeptuneLogger
from torch.utils.data import DataLoader, Dataset

import neptune.new as neptune


class RandomDataset(Dataset):
    def __init__(self, size, length):
        self.len = length
        # pylint: disable=no-member
        self.data = torch.randn(length, size)

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return self.len


class BoringModel(LightningModule):
    # pylint: disable=abstract-method
    def __init__(self):
        super().__init__()
        self.layer = torch.nn.Linear(32, 2)

    def forward(self, *args, **kwargs):
        return self.layer(*args, **kwargs)

    def training_step(self, *args, **kwargs):
        batch, *_ = args
        loss = self(batch).sum()
        self.log("train/loss", loss)
        return {"loss": loss}

    def validation_step(self, *args, **kwargs):
        batch, *_ = args
        loss = self(batch, **kwargs).sum()
        self.log("valid/loss", loss)

    def test_step(self, *args, **kwargs):
        batch, *_ = args
        loss = self(batch, **kwargs).sum()
        self.log("test/loss", loss)

    def configure_optimizers(self):
        return torch.optim.SGD(self.layer.parameters(), lr=0.1)


@pytest.fixture(scope="session")
def pytorch_run(environment):
    # given
    run = neptune.init(name="Pytorch-Lightning integration", project=environment.project)
    # and
    model_checkpoint = ModelCheckpoint(
        dirpath="my_model/checkpoints/",
        filename="{epoch:02d}-{valid/loss:.2f}",
        save_weights_only=True,
        save_top_k=2,
        save_last=True,
        monitor="valid/loss",
        every_n_epochs=1,
    )
    neptune_logger = NeptuneLogger(run=run, prefix="custom_prefix")
    # and (Subject)
    model = BoringModel()
    trainer = Trainer(
        limit_train_batches=1,
        limit_val_batches=1,
        log_every_n_steps=1,
        max_epochs=3,
        logger=neptune_logger,
        callbacks=[model_checkpoint],
    )
    train_data = DataLoader(RandomDataset(32, 64), batch_size=2)
    val_data = DataLoader(RandomDataset(32, 64), batch_size=2)
    test_data = DataLoader(RandomDataset(32, 64), batch_size=2)

    # then
    trainer.fit(model, train_dataloaders=train_data, val_dataloaders=val_data)
    trainer.test(model, dataloaders=test_data)
    run.sync()

    yield run



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
from unittest.mock import MagicMock

import pytest

from neptune.new.exceptions import ProjectNotFound
from neptune.new.sync.utils import get_project


@pytest.fixture(name="backend")
def backend_fixture():
    return MagicMock()


def test_get_project_no_name_set(mocker, backend):
    # given
    mocker.patch.object(os, "getenv")
    os.getenv.return_value = None

    # expect
    assert get_project(None, backend=backend) is None


def test_get_project_project_not_found(backend):
    # given
    backend.get_project.side_effect = ProjectNotFound("foo")

    # expect
    assert get_project("foo", backend=backend) is None



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest
from unittest.mock import MagicMock

from neptune.new.internal.backends.swagger_client_wrapper import SwaggerClientWrapper


class TestSwaggerClientWrapper(unittest.TestCase):
    # pylint:disable=protected-access

    def setUp(self) -> None:
        pass

    def test_api_callable_objects(self):
        # given
        swagger_client = MagicMock()
        api = MagicMock()
        api.callable_object = MagicMock()
        api.callable_object.sub_property = 13
        swagger_client.api = api

        wrapper = SwaggerClientWrapper(swagger_client)

        # when
        wrapper.api.method("arg1", kwarg="kwarg1")
        wrapper.api.callable_object("arg2", kwarg="kwarg2")

        # then
        api.method.assert_called_once_with("arg1", kwarg="kwarg1")
        api.callable_object.assert_called_once_with("arg2", kwarg="kwarg2")
        self.assertEqual(13, wrapper.api.callable_object.sub_property)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest
import uuid

from neptune.new.exceptions import MetadataInconsistency
from neptune.new.internal.backends.neptune_backend_mock import NeptuneBackendMock
from neptune.new.internal.container_structure import ContainerStructure
from neptune.new.internal.container_type import ContainerType
from neptune.new.types.value import Value


class TestRunStructure(unittest.TestCase):
    def test_get_none(self):
        exp = ContainerStructure[int, dict]()
        self.assertEqual(exp.get(["some", "path", "val"]), None)

    def test_get_nested_variable_fails(self):
        exp = ContainerStructure[int, dict]()
        exp.set(["some", "path", "val"], 3)
        with self.assertRaises(MetadataInconsistency):
            exp.get(["some", "path", "val", "nested"])
        with self.assertRaises(MetadataInconsistency):
            exp.get(["some", "path", "val", "nested", "nested"])

    def test_get_ns(self):
        exp = ContainerStructure[int, dict]()
        exp.set(["some", "path", "val"], 3)
        self.assertEqual(exp.get(["some", "path"]), {"val": 3})

    def test_set(self):
        exp = ContainerStructure[int, dict]()
        exp.set(["some", "path", "val"], 3)
        self.assertEqual(exp.get(["some", "path", "val"]), 3)

    def test_set_nested_variable_fails(self):
        exp = ContainerStructure[int, dict]()
        exp.set(["some", "path", "val"], 3)
        with self.assertRaises(MetadataInconsistency):
            exp.set(["some", "path", "val", "nested"], 3)
        with self.assertRaises(MetadataInconsistency):
            exp.set(["some", "path", "val", "nested", "nested"], 3)

    def test_set_ns_collision(self):
        exp = ContainerStructure[int, dict]()
        exp.set(["some", "path", "val"], 3)
        with self.assertRaises(MetadataInconsistency):
            exp.set(["some", "path"], 5)

    def test_pop(self):
        exp = ContainerStructure[int, dict]()
        exp.set(["some", "path", "val1"], 3)
        exp.set(["some", "path", "val2"], 5)
        exp.pop(["some", "path", "val2"])
        self.assertEqual(exp.get(["some", "path", "val1"]), 3)
        self.assertEqual(exp.get(["some", "path", "val2"]), None)
        self.assertTrue("some" in exp.get_structure() and "path" in exp.get_structure()["some"])

    def test_pop_whole_ns(self):
        exp = ContainerStructure[int, dict]()
        exp.set(["some", "path", "val"], 3)
        exp.pop(["some", "path", "val"])
        self.assertEqual(exp.get(["some", "path", "val"]), None)
        self.assertFalse("some" in exp.get_structure())

    def test_pop_not_found(self):
        exp = ContainerStructure[int, dict]()
        with self.assertRaises(MetadataInconsistency):
            exp.pop(["some", "path"])

    def test_pop_ns_fail(self):
        exp = ContainerStructure[int, dict]()
        exp.set(["some", "path", "val1"], 3)
        with self.assertRaises(MetadataInconsistency):
            exp.pop(["some", "path"])


class TestIterateSubpaths(unittest.TestCase):
    # pylint: disable=protected-access
    project_uuid = str(uuid.uuid4())

    def setUp(self):
        self.backend = NeptuneBackendMock()
        exp = self.backend.create_run(self.project_uuid)
        # FIXME test for projects
        self.structure = self.backend._containers[(exp.id, ContainerType.RUN)]
        self.structure.set(["attributes", "float"], Value())
        self.structure.set(["attributes", "node", "one"], Value())
        self.structure.set(["attributes", "node", "two"], Value())
        self.structure.set(["attributes", "node", "three"], Value())
        self.structure.set(["attributes", "int"], Value())
        self.structure.set(["attributes", "string"], Value())

    def test_iterate_empty_run(self):
        empty_structure = ContainerStructure[Value, dict]()

        self.assertListEqual(list(empty_structure.iterate_subpaths([])), [])
        self.assertListEqual(list(empty_structure.iterate_subpaths(["test"])), [])

    def test_iterate_empty_prefix(self):
        prefix = []
        expected_subpaths = [
            "sys/id",
            "sys/state",
            "sys/owner",
            "sys/size",
            "sys/tags",
            "sys/creation_time",
            "sys/modification_time",
            "sys/failed",
            "attributes/float",
            "attributes/int",
            "attributes/string",
            "attributes/node/one",
            "attributes/node/two",
            "attributes/node/three",
        ]

        self.assertListEqual(list(self.structure.iterate_subpaths(prefix)), expected_subpaths)

    def test_iterate_prefix(self):
        prefix = ["sys"]
        expected_subpaths = [
            "sys/id",
            "sys/state",
            "sys/owner",
            "sys/size",
            "sys/tags",
            "sys/creation_time",
            "sys/modification_time",
            "sys/failed",
        ]

        self.assertListEqual(list(self.structure.iterate_subpaths(prefix)), expected_subpaths)

    def test_iterate_long_prefix(self):
        prefix = ["attributes", "node"]
        expected_subpaths = [
            "attributes/node/one",
            "attributes/node/two",
            "attributes/node/three",
        ]

        self.assertListEqual(list(self.structure.iterate_subpaths(prefix)), expected_subpaths)

    def test_iterate_nonexistent_prefix(self):
        prefix = ["argh"]
        expected_subpaths = []

        self.assertListEqual(list(self.structure.iterate_subpaths(prefix)), expected_subpaths)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from unittest.mock import MagicMock

import pytest

from neptune.new.internal.container_type import ContainerType
from neptune.new.internal.operation import Operation
from neptune.new.sync import StatusRunner
from neptune.new.sync.utils import get_qualified_name
from tests.neptune.new.sync.utils import (
    generate_get_metadata_container,
    prepare_metadata_container,
)


@pytest.fixture(name="backend")
def backend_fixture():
    return MagicMock()


@pytest.fixture(name="status_runner")
def status_runner_fixture(backend):
    return StatusRunner(backend=backend)


@pytest.mark.parametrize("container_type", list(ContainerType))
def test_list_containers(tmp_path, mocker, capsys, backend, status_runner, container_type):
    # given
    unsynced_container = prepare_metadata_container(
        container_type=container_type, path=tmp_path, last_ack_version=1
    )
    synced_container = prepare_metadata_container(
        container_type=container_type, path=tmp_path, last_ack_version=3
    )
    get_container_impl = generate_get_metadata_container(
        registered_containers=(unsynced_container, synced_container)
    )

    # and
    mocker.patch.object(backend, "get_metadata_container", get_container_impl)
    mocker.patch.object(Operation, "from_dict")

    # when
    status_runner.synchronization_status(tmp_path)

    # then
    captured = capsys.readouterr()
    assert captured.out.splitlines() == [
        "Unsynchronized objects:",
        f"- {get_qualified_name(unsynced_container)}",
        "Synchronized objects:",
        f"- {get_qualified_name(synced_container)}",
        "",
        "Please run with the `neptune sync --help` to see example commands.",
    ]


def test_list_offline_runs(tmp_path, mocker, capsys, status_runner):
    # given
    offline_run = prepare_metadata_container(
        container_type=ContainerType.RUN,
        path=tmp_path,
        last_ack_version=None,
    )

    # and
    mocker.patch.object(Operation, "from_dict")

    # when
    status_runner.synchronization_status(tmp_path)

    # then
    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        "Unsynchronized offline objects:\n- offline/run__{}".format(offline_run.id) in captured.out
    )


def test_list_runs_when_no_run(tmp_path, capsys, status_runner):
    (tmp_path / "async").mkdir()
    # when
    with pytest.raises(SystemExit):
        status_runner.synchronization_status(tmp_path)

    # then
    captured = capsys.readouterr()
    assert captured.err == ""
    assert "There are no Neptune objects" in captured.out



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
import unittest
from io import BytesIO

import pytest
from mock import Mock, patch

from neptune.internal.api_clients.client_config import MultipartConfig
from neptune.internal.storage.datastream import FileChunk, FileChunker, FileChunkStream
from neptune.internal.storage.storage_utils import (
    AttributeUploadConfiguration,
    UploadEntry,
)
from neptune.new.exceptions import InternalClientError


class TestFileChunkStream(unittest.TestCase):
    @patch("os.path.exists", new=lambda _: True)
    @patch("stat.S_ISDIR", new=lambda _: False)
    @patch("os.lstat")
    def test_permissions_to_unix_string_for_file(self, lstat):
        # given
        lstat.return_value.st_mode = 0o731

        # when
        permissions_string = UploadEntry.permissions_to_unix_string("/some/path")

        # then
        self.assertEqual("-rwx-wx--x", permissions_string)

    @patch("os.path.exists", new=lambda _: True)
    @patch("stat.S_ISDIR", new=lambda _: True)
    @patch("os.lstat")
    def test_permissions_to_unix_string_for_directory(self, lstat):
        # given
        lstat.return_value.st_mode = 0o642

        # when
        permissions_string = UploadEntry.permissions_to_unix_string("/some/path")

        # then
        self.assertEqual("drw-r---w-", permissions_string)

    @patch("os.path.exists", new=lambda _: False)
    def test_permissions_to_unix_string_for_nonexistent_file(self):
        # when
        permissions_string = UploadEntry.permissions_to_unix_string("/some/path")

        # then
        self.assertEqual("-" * 10, permissions_string)

    def test_generate_chunks_from_stream(self):
        # given
        text = "ABCDEFGHIJKLMNOPRSTUWXYZ"

        # when
        stream = FileChunkStream(
            UploadEntry(BytesIO(bytes(text, "utf-8")), "some/path"),
            AttributeUploadConfiguration(10),
        )
        chunks = list()
        for chunk in stream.generate():
            chunks.append(chunk)

        # then
        self.assertEqual(stream.length, 24)
        self.assertEqual(
            chunks,
            [
                FileChunk(b"ABCDEFGHIJ", 0, 10),
                FileChunk(b"KLMNOPRSTU", 10, 20),
                FileChunk(b"WXYZ", 20, 24),
            ],
        )


class TestFileChunker:
    multipart_config = MultipartConfig(
        min_chunk_size=5_242_880,  # 5 MB
        max_chunk_size=1_073_741_824,  # 1 GB
        max_chunk_count=1_000,
        max_single_part_size=5_242_880,  # 1 GB
    )

    def get_chunk_count(self, file_size, chunk_size):
        chunk_idx = 0
        while file_size > chunk_size:
            chunk_idx += 1
            file_size -= chunk_size
        return chunk_idx + 1

    @pytest.mark.parametrize(
        "file_size, expected_chunk_size, expected_chunk_count",
        (
            (1_000_000, 5_242_880, 1),
            (6_000_000, 5_242_880, 2),
            (5_242_880_000, 5_242_880, 1_000),
            (5_242_880_001, 5_242_881, 1_000),
            (5_242_891_001, 5_242_892, 1_000),
            (1_073_741_824_000, 1_073_741_824, 1_000),
        ),
    )
    def test_chunk_size_for_small_file(self, file_size, expected_chunk_size, expected_chunk_count):
        chunker = FileChunker(
            Mock(), Mock(), total_size=file_size, multipart_config=self.multipart_config
        )

        chunk_size = chunker._get_chunk_size()

        chunk_count = self.get_chunk_count(file_size, chunk_size)
        assert chunk_count == expected_chunk_count
        assert chunk_size == expected_chunk_size
        assert chunk_count <= self.multipart_config.max_chunk_count
        assert chunk_size <= self.multipart_config.max_chunk_size
        assert chunk_size >= self.multipart_config.min_chunk_size

    def test_too_large_file(self):
        file_size = 1_073_741_824_001

        chunker = FileChunker(
            Mock(), Mock(), total_size=file_size, multipart_config=self.multipart_config
        )

        with pytest.raises(InternalClientError):
            chunker._get_chunk_size()


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from neptune.new.internal.backends.nql import (
    NQLAttributeOperator,
    NQLAttributeType,
    NQLQueryAttribute,
)
from neptune.new.internal.container_type import ContainerType
from neptune.new.metadata_containers import MetadataContainer
from neptune.new.metadata_containers.metadata_containers_table import Table


class Model(MetadataContainer):
    """A class for managing a Neptune model and retrieving information from it.

    You may also want to check `Model docs page`_.

    .. _Model docs page:
       https://docs.neptune.ai/api-reference/model
    """

    container_type = ContainerType.MODEL

    @property
    def _docs_url_stop(self) -> str:
        return "https://docs.neptune.ai/api-reference/model#.stop"

    @property
    def _label(self) -> str:
        return self._sys_id

    @property
    def _url(self) -> str:
        return self._backend.get_model_url(
            model_id=self._id,
            workspace=self._workspace,
            project_name=self._project_name,
            sys_id=self._sys_id,
        )

    @property
    def _metadata_url(self) -> str:
        return self._url.rstrip("/") + "/metadata"

    def fetch_model_versions_table(self) -> Table:
        """Retrieve all model versions of the given model.

        Returns:
            ``Table``: object containing experiments matching the specified criteria.

            Use `.to_pandas()` to convert it to Pandas `DataFrame`.
        """
        return MetadataContainer._fetch_entries(
            self,
            child_type=ContainerType.MODEL_VERSION,
            query=NQLQueryAttribute(
                name="sys/model_id",
                value=self._sys_id,
                operator=NQLAttributeOperator.EQUALS,
                type=NQLAttributeType.STRING,
            ),
        )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = []

from pathlib import Path
from typing import List, Optional

import click

from neptune.new.constants import NEPTUNE_DATA_DIRECTORY
from neptune.new.exceptions import (
    CannotSynchronizeOfflineRunsWithoutProject,
    NeptuneConnectionLostException,
    NeptuneException,
    ProjectNotFound,
    RunNotFound,
)
from neptune.new.internal.backends.api_model import ApiExperiment, Project
from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.credentials import Credentials
from neptune.new.internal.disk_queue import DiskQueue
from neptune.new.internal.operation import Operation
from neptune.new.internal.utils.logger import logger
from neptune.new.sync.status import StatusRunner
from neptune.new.sync.sync import SyncRunner

# backwards compatibility
# pylint: disable=unused-import,wrong-import-order


# pylint: disable=unused-argument
def get_neptune_path(ctx, param, path: str) -> Path:
    # check if path exists and contains a '.neptune' folder
    path = Path(path)
    if (path / NEPTUNE_DATA_DIRECTORY).is_dir():
        return path / NEPTUNE_DATA_DIRECTORY
    elif path.name == NEPTUNE_DATA_DIRECTORY and path.is_dir():
        return path
    else:
        raise click.BadParameter(
            "Path {} does not contain a '{}' folder.".format(path, NEPTUNE_DATA_DIRECTORY)
        )


path_option = click.option(
    "--path",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    default=Path.cwd(),
    callback=get_neptune_path,
    metavar="<location>",
    help="path to a directory containing a '.neptune' folder with stored objects",
)


@click.command()
@path_option
def status(path: Path) -> None:
    """List synchronized and unsynchronized objects in the given directory. Trashed objects are not listed.

    Neptune stores object data on disk in '.neptune' directories. In case an object executes offline
    or network is unavailable as the object executes, object data can be synchronized
    with the server with this command line utility.

    Examples:

    \b
    # List synchronized and unsynchronized objects in the current directory
    neptune status

    \b
    # List synchronized and unsynchronized objects in directory "foo/bar" without actually syncing
    neptune status --path foo/bar
    """

    status_runner = StatusRunner(backend=HostedNeptuneBackend(Credentials.from_token()))

    status_runner.synchronization_status(path)


@click.command()
@path_option
@click.option(
    "--run",
    "runs_names",
    multiple=True,
    metavar="<run-name>",
    help="[deprecated] run name (workspace/project/short-id or UUID for offline runs) to synchronize.",
)
@click.option(
    "--object",
    "object_names",
    multiple=True,
    metavar="<object-name>",
    help="object name (workspace/project/short-id or UUID for offline runs) to synchronize.",
)
@click.option(
    "-p",
    "--project",
    "project_name",
    multiple=False,
    metavar="project-name",
    help="project name (workspace/project) where offline runs will be sent",
)
@click.option(
    "--offline-only",
    "offline_only",
    is_flag=True,
    default=False,
    help="synchronize only the offline runs inside '.neptune' directory",
)
def sync(
    path: Path,
    runs_names: List[str],
    object_names: List[str],
    project_name: Optional[str],
    offline_only: Optional[bool],
):
    """Synchronizes objects with unsent data with the server.

    Neptune stores object data on disk in '.neptune' directories. In case a object executes offline
    or network is unavailable as the run executes, object data can be synchronized
    with the server with this command line utility.

    You can list unsynchronized runs with `neptune status`

    Examples:

    \b
    # Synchronize all objects in the current directory
    neptune sync

    \b
    # Synchronize all objects in the given path
    neptune sync --path foo/bar

    \b
    # Synchronize only runs "NPT-42" and "NPT-43" in "workspace/project" in the current directory
    neptune sync --object workspace/project/NPT-42 --object workspace/project/NPT-43

    \b
    # Synchronise all objects in the current directory, sending offline runs to project "workspace/project"
    neptune sync --project workspace/project

    \b
    # Synchronize only the offline run with UUID offline/a1561719-b425-4000-a65a-b5efb044d6bb
    # to project "workspace/project"
    neptune sync --project workspace/project --object offline/a1561719-b425-4000-a65a-b5efb044d6bb

    \b
    # Synchronize only the offline runs
    neptune sync --offline-only

    \b
    # Synchronize only the offline runs to project "workspace/project"
    neptune sync --project workspace/project --offline-only
    """

    sync_runner = SyncRunner(backend=HostedNeptuneBackend(Credentials.from_token()))

    if runs_names:
        logger.warning(
            "WARNING: --run parameter is deprecated and will be removed in the future, please start using --object"
        )
        # prefer object_names, obviously
        object_names = set(object_names)
        object_names.update(runs_names)

    if offline_only:
        if object_names:
            raise click.BadParameter("--object and --offline-only are mutually exclusive")

        sync_runner.sync_all_offline_containers(path, project_name)

    elif object_names:
        sync_runner.sync_selected_containers(path, project_name, object_names)
    else:
        sync_runner.sync_all_containers(path, project_name)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import collections
import enum
import json
import os
import time
from io import BytesIO
from typing import AnyStr, Dict, Iterable, List, Optional, Set, Union
from urllib.parse import urlencode

from bravado.exception import HTTPPaymentRequired, HTTPUnprocessableEntity
from bravado.requests_client import RequestsClient
from requests import Request, Response

from neptune.internal.hardware.constants import BYTES_IN_ONE_MB
from neptune.internal.storage.datastream import (
    FileChunk,
    FileChunker,
    FileChunkStream,
    compress_to_tar_gz_in_memory,
)
from neptune.internal.storage.storage_utils import (
    AttributeUploadConfiguration,
    UploadEntry,
    normalize_file_name,
    scan_unique_upload_entries,
    split_upload_files,
)
from neptune.new.exceptions import (
    FileUploadError,
    InternalClientError,
    MetadataInconsistency,
    NeptuneException,
    NeptuneLimitExceedException,
    UploadedFileChanged,
)
from neptune.new.internal.backends.api_model import MultipartConfig
from neptune.new.internal.backends.swagger_client_wrapper import (
    ApiMethodWrapper,
    SwaggerClientWrapper,
)
from neptune.new.internal.backends.utils import (
    build_operation_url,
    handle_server_raw_response_messages,
    with_api_exceptions_handler,
)
from neptune.new.internal.utils import get_absolute_paths, get_common_root
from neptune.new.internal.utils.logger import logger

DEFAULT_CHUNK_SIZE = 5 * BYTES_IN_ONE_MB
DEFAULT_UPLOAD_CONFIG = AttributeUploadConfiguration(chunk_size=DEFAULT_CHUNK_SIZE)


class FileUploadTarget(enum.Enum):
    FILE_ATOM = "file"
    FILE_SET = "fileset"


def upload_file_attribute(
    swagger_client: SwaggerClientWrapper,
    container_id: str,
    attribute: str,
    source: Union[str, bytes],
    ext: str,
    multipart_config: Optional[MultipartConfig],
) -> List[NeptuneException]:
    if isinstance(source, str) and not os.path.isfile(source):
        return [FileUploadError(source, "Path not found or is a not a file.")]

    target = attribute
    if ext:
        target += "." + ext

    try:
        upload_entry = UploadEntry(source if isinstance(source, str) else BytesIO(source), target)
        if multipart_config is None:
            # the legacy upload procedure
            url = build_operation_url(
                swagger_client.swagger_spec.api_url,
                swagger_client.api.uploadAttribute.operation.path_name,
            )
            upload_configuration = DEFAULT_UPLOAD_CONFIG

            _upload_loop(
                file_chunk_stream=FileChunkStream(upload_entry, upload_configuration),
                http_client=swagger_client.swagger_spec.http_client,
                url=url,
                query_params={
                    "experimentId": container_id,
                    "attribute": attribute,
                    "ext": ext,
                },
            )
        else:
            _multichunk_upload_with_retry(
                upload_entry,
                query_params={
                    "experimentIdentifier": container_id,
                    "attribute": attribute,
                    "ext": ext,
                },
                swagger_client=swagger_client,
                multipart_config=multipart_config,
                target=FileUploadTarget.FILE_ATOM,
            )
    except MetadataInconsistency as e:
        return [e]


def upload_file_set_attribute(
    swagger_client: SwaggerClientWrapper,
    container_id: str,
    attribute: str,
    file_globs: Iterable[str],
    reset: bool,
    multipart_config: Optional[MultipartConfig],
) -> List[NeptuneException]:
    unique_upload_entries = get_unique_upload_entries(file_globs)

    try:
        upload_configuration = DEFAULT_UPLOAD_CONFIG
        for package in split_upload_files(
            upload_entries=unique_upload_entries,
            upload_configuration=upload_configuration,
        ):
            if package.is_empty() and not reset:
                continue

            uploading_multiple_entries = package.len > 1
            creating_a_single_empty_dir = (
                package.len == 1
                and not package.items[0].is_stream()
                and os.path.isdir(package.items[0].source)
            )

            if uploading_multiple_entries or creating_a_single_empty_dir or package.is_empty():
                data = compress_to_tar_gz_in_memory(upload_entries=package.items)
                url = build_operation_url(
                    swagger_client.swagger_spec.api_url,
                    swagger_client.api.uploadFileSetAttributeTar.operation.path_name,
                )
                result = upload_raw_data(
                    http_client=swagger_client.swagger_spec.http_client,
                    url=url,
                    data=data,
                    headers={"Content-Type": "application/octet-stream"},
                    query_params={
                        "experimentId": container_id,
                        "attribute": attribute,
                        "reset": str(reset),
                    },
                )
                _attribute_upload_response_handler(result)
            else:
                upload_entry = package.items[0]
                if multipart_config is None:
                    # the legacy upload procedure
                    url = build_operation_url(
                        swagger_client.swagger_spec.api_url,
                        swagger_client.api.uploadFileSetAttributeChunk.operation.path_name,
                    )
                    file_chunk_stream = FileChunkStream(
                        upload_entry=upload_entry,
                        upload_configuration=upload_configuration,
                    )
                    _upload_loop(
                        file_chunk_stream=file_chunk_stream,
                        http_client=swagger_client.swagger_spec.http_client,
                        url=url,
                        query_params={
                            "experimentId": container_id,
                            "attribute": attribute,
                            "reset": str(reset),
                            "path": upload_entry.target_path,
                        },
                    )
                else:
                    _multichunk_upload_with_retry(
                        upload_entry,
                        query_params={
                            "experimentIdentifier": container_id,
                            "attribute": attribute,
                            "subPath": upload_entry.target_path,
                        },
                        swagger_client=swagger_client,
                        multipart_config=multipart_config,
                        target=FileUploadTarget.FILE_SET,
                    )

            reset = False
    except MetadataInconsistency as e:
        if len(e.args) == 1:
            return [e]
        else:
            return [MetadataInconsistency(desc) for desc in e.args]


def get_unique_upload_entries(file_globs: Iterable[str]) -> Set[UploadEntry]:
    absolute_paths = get_absolute_paths(file_globs)
    common_root = get_common_root(absolute_paths)

    upload_entries: List[UploadEntry] = []
    if common_root is not None:
        for absolute_path in absolute_paths:
            upload_entries.append(
                UploadEntry(
                    absolute_path,
                    normalize_file_name(os.path.relpath(absolute_path, common_root)),
                )
            )
    else:
        for absolute_path in absolute_paths:
            upload_entries.append(UploadEntry(absolute_path, normalize_file_name(absolute_path)))

    return scan_unique_upload_entries(upload_entries)


def _attribute_upload_response_handler(result: bytes) -> None:
    try:
        parsed = json.loads(result)
    except json.JSONDecodeError:
        raise InternalClientError("Unexpected response from server: {}".format(result))

    if isinstance(parsed, type(None)):
        # old format with empty optional error
        return
    if isinstance(parsed, dict):
        if "errorDescription" in parsed:
            # old format with optional error
            raise MetadataInconsistency(parsed["errorDescription"])
        elif "errors" in parsed:
            # new format with a list of errors
            error_list = parsed["errors"]
            if isinstance(error_list, list):
                if len(error_list) == 0:
                    return
                try:
                    raise MetadataInconsistency(
                        *[item["errorDescription"] for item in parsed["errors"]]
                    )
                except KeyError:
                    # fall into default InternalClientError
                    pass

    raise InternalClientError("Unexpected response from server: {}".format(result))


MultipartUrlSet = collections.namedtuple(
    "MultipartUrlSet", ["start_chunked", "finish_chunked", "send_chunk", "single"]
)

MULTIPART_URLS = {
    FileUploadTarget.FILE_ATOM: MultipartUrlSet(
        "fileAtomMultipartUploadStart",
        "fileAtomMultipartUploadFinish",
        "fileAtomMultipartUploadPart",
        "fileAtomUpload",
    ),
    FileUploadTarget.FILE_SET: MultipartUrlSet(
        "fileSetFileMultipartUploadStart",
        "fileSetFileMultipartUploadFinish",
        "fileSetFileMultipartUploadPart",
        "fileSetFileUpload",
    ),
}


def _build_multipart_urlset(
    swagger_client: SwaggerClientWrapper, target: FileUploadTarget
) -> MultipartUrlSet:
    urlnameset = MULTIPART_URLS[target]
    return MultipartUrlSet(
        start_chunked=with_api_exceptions_handler(
            getattr(swagger_client.api, urlnameset.start_chunked)
        ),
        finish_chunked=with_api_exceptions_handler(
            getattr(swagger_client.api, urlnameset.finish_chunked)
        ),
        send_chunk=build_operation_url(
            swagger_client.swagger_spec.api_url,
            getattr(swagger_client.api, urlnameset.send_chunk).operation.path_name,
        ),
        single=build_operation_url(
            swagger_client.swagger_spec.api_url,
            getattr(swagger_client.api, urlnameset.single).operation.path_name,
        ),
    )


def _multichunk_upload_with_retry(
    upload_entry: UploadEntry,
    swagger_client: SwaggerClientWrapper,
    query_params: dict,
    multipart_config: MultipartConfig,
    target: FileUploadTarget,
):
    urlset = _build_multipart_urlset(swagger_client, target)
    while True:
        try:
            return _multichunk_upload(
                upload_entry, swagger_client, query_params, multipart_config, urlset
            )
        except UploadedFileChanged as e:
            logger.error(str(e))


def _multichunk_upload(
    upload_entry: UploadEntry,
    swagger_client: SwaggerClientWrapper,
    query_params: dict,
    multipart_config: MultipartConfig,
    urlset: MultipartUrlSet,
):
    file_stream = upload_entry.get_stream()
    entry_length = upload_entry.length()
    try:
        if entry_length <= multipart_config.max_single_part_size:
            # single upload
            data = file_stream.read()
            result = upload_raw_data(
                http_client=swagger_client.swagger_spec.http_client,
                url=urlset.single,
                data=data,
                query_params=query_params,
            )
            _attribute_upload_response_handler(result)
        else:
            # chunked upload
            result = (
                urlset.start_chunked(**query_params, totalLength=entry_length).response().result
            )
            if result.errors:
                raise MetadataInconsistency([err.errorDescription for err in result.errors])

            no_ext_query_params = query_params.copy()
            if "ext" in no_ext_query_params:
                del no_ext_query_params["ext"]

            upload_id = result.uploadId
            chunker = FileChunker(
                None if upload_entry.is_stream() else upload_entry.source,
                file_stream,
                entry_length,
                multipart_config,
            )
            for idx, chunk in enumerate(chunker.generate()):
                result = upload_raw_data(
                    http_client=swagger_client.swagger_spec.http_client,
                    url=urlset.send_chunk,
                    data=chunk.data,
                    headers={"X-Range": _build_x_range(chunk, entry_length)},
                    query_params={
                        "uploadId": upload_id,
                        "uploadPartIdx": idx,
                        **no_ext_query_params,
                    },
                )
                _attribute_upload_response_handler(result)

            result = (
                urlset.finish_chunked(**no_ext_query_params, uploadId=upload_id).response().result
            )
            if result.errors:
                raise MetadataInconsistency([err.errorDescription for err in result.errors])
        return []
    finally:
        file_stream.close()


def _upload_loop(file_chunk_stream: FileChunkStream, query_params: dict, **kwargs):
    for iteration, chunk in enumerate(file_chunk_stream.generate()):
        if "reset" in query_params and iteration != 0:
            query_params["reset"] = str(False)
        result = _upload_loop_chunk(
            chunk, file_chunk_stream, query_params=query_params.copy(), **kwargs
        )
        _attribute_upload_response_handler(result)

    file_chunk_stream.close()


def _build_x_range(chunk: FileChunk, total_size: int) -> str:
    return "bytes=%d-%d/%d" % (
        chunk.start,
        chunk.end - 1,
        total_size,
    )


def _upload_loop_chunk(
    chunk: FileChunk, file_chunk_stream: FileChunkStream, query_params: dict, **kwargs
):
    binary_range = _build_x_range(chunk, file_chunk_stream.length)
    headers = {
        "Content-Filename": file_chunk_stream.filename,
        "X-Range": binary_range,
        "Content-Type": "application/octet-stream",
    }
    if file_chunk_stream.permissions is not None:
        headers["X-File-Permissions"] = file_chunk_stream.permissions
    return upload_raw_data(data=chunk.data, headers=headers, query_params=query_params, **kwargs)


@with_api_exceptions_handler
def upload_raw_data(
    http_client: RequestsClient,
    url: str,
    data: AnyStr,
    path_params: Optional[Dict[str, str]] = None,
    query_params: Optional[Dict[str, str]] = None,
    headers: Optional[Dict[str, str]] = None,
):
    url = _generate_url(url=url, path_params=path_params, query_params=query_params)

    session = http_client.session
    request = http_client.authenticator.apply(
        Request(method="POST", url=url, data=data, headers=headers)
    )
    response = handle_server_raw_response_messages(session.send(session.prepare_request(request)))

    if response.status_code >= 300:
        ApiMethodWrapper.handle_neptune_http_errors(response)
    if response.status_code in (
        HTTPUnprocessableEntity.status_code,
        HTTPPaymentRequired.status_code,
    ):
        raise NeptuneLimitExceedException(reason=response.json().get("title", "Unknown reason"))
    response.raise_for_status()
    return response.content


def download_image_series_element(
    swagger_client: SwaggerClientWrapper,
    container_id: str,
    attribute: str,
    index: int,
    destination: str,
):
    url = build_operation_url(
        swagger_client.swagger_spec.api_url,
        swagger_client.api.getImageSeriesValue.operation.path_name,
    )
    response = _download_raw_data(
        http_client=swagger_client.swagger_spec.http_client,
        url=url,
        headers={},
        query_params={
            "experimentId": container_id,
            "attribute": attribute,
            "index": index,
        },
    )
    _store_response_as_file(
        response,
        os.path.join(
            destination,
            "{}.{}".format(index, response.headers["content-type"].split("/")[-1]),
        ),
    )


def download_file_attribute(
    swagger_client: SwaggerClientWrapper,
    container_id: str,
    attribute: str,
    destination: Optional[str] = None,
):
    url = build_operation_url(
        swagger_client.swagger_spec.api_url,
        swagger_client.api.downloadAttribute.operation.path_name,
    )
    response = _download_raw_data(
        http_client=swagger_client.swagger_spec.http_client,
        url=url,
        headers={"Accept": "application/octet-stream"},
        query_params={"experimentId": container_id, "attribute": attribute},
    )
    _store_response_as_file(response, destination)


def download_file_set_attribute(
    swagger_client: SwaggerClientWrapper,
    download_id: str,
    destination: Optional[str] = None,
):
    download_url: Optional[str] = _get_download_url(swagger_client, download_id)
    next_sleep = 0.5
    while download_url is None:
        time.sleep(next_sleep)
        next_sleep = min(2 * next_sleep, 5)
        download_url = _get_download_url(swagger_client, download_id)

    response = _download_raw_data(
        http_client=swagger_client.swagger_spec.http_client,
        url=download_url,
        headers={"Accept": "application/zip"},
    )
    _store_response_as_file(response, destination)


def _get_download_url(swagger_client: SwaggerClientWrapper, download_id: str):
    params = {"id": download_id}
    download_request = swagger_client.api.getDownloadPrepareRequest(**params).response().result
    return download_request.downloadUrl


def _store_response_as_file(response: Response, destination: Optional[str] = None):
    if destination is None:
        target_file = _get_content_disposition_filename(response)
    elif os.path.isdir(destination):
        target_file = os.path.join(destination, _get_content_disposition_filename(response))
    else:
        target_file = destination
    with response:
        with open(target_file, "wb") as f:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                if chunk:
                    f.write(chunk)


def _get_content_disposition_filename(response: Response) -> str:
    content_disposition = response.headers["Content-Disposition"]
    return content_disposition[content_disposition.rfind("filename=") + 9 :].strip('"')


@with_api_exceptions_handler
def _download_raw_data(
    http_client: RequestsClient,
    url: str,
    path_params: Optional[Dict[str, str]] = None,
    query_params: Optional[Dict[str, str]] = None,
    headers: Optional[Dict[str, str]] = None,
) -> Response:
    url = _generate_url(url=url, path_params=path_params, query_params=query_params)

    session = http_client.session
    request = http_client.authenticator.apply(Request(method="GET", url=url, headers=headers))

    response = handle_server_raw_response_messages(
        session.send(session.prepare_request(request), stream=True)
    )

    response.raise_for_status()
    return response


def _generate_url(
    url: str,
    path_params: Optional[Dict[str, str]] = None,
    query_params: Optional[Dict[str, str]] = None,
) -> str:
    for key, val in (path_params or dict()).items():
        url = url.replace("{" + key + "}", val)
    if query_params:
        url = url + "?" + urlencode(list(query_params.items()))
    return url



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mock import MagicMock


class BackendTestMixin:
    @staticmethod
    def _get_swagger_client_mock(
        swagger_client_factory,
        min_recommended=None,
        min_compatible=None,
        max_compatible=None,
    ):
        py_lib_versions = type("py_lib_versions", (object,), {})()
        setattr(py_lib_versions, "minRecommendedVersion", min_recommended)
        setattr(py_lib_versions, "minCompatibleVersion", min_compatible)
        setattr(py_lib_versions, "maxCompatibleVersion", max_compatible)

        artifacts = type("artifacts", (object,), {})()
        setattr(artifacts, "enabled", True)

        multipart_upload = type("multiPartUpload", (object,), {})()
        setattr(multipart_upload, "enabled", True)
        setattr(multipart_upload, "minChunkSize", 5242880)
        setattr(multipart_upload, "maxChunkSize", 1073741824)
        setattr(multipart_upload, "maxChunkCount", 1000)
        setattr(multipart_upload, "maxSinglePartSize", 5242880)

        client_config = type("client_config_response_result", (object,), {})()
        setattr(client_config, "pyLibVersions", py_lib_versions)
        setattr(client_config, "artifacts", artifacts)
        setattr(client_config, "multiPartUpload", multipart_upload)
        setattr(client_config, "apiUrl", "ui.neptune.ai")
        setattr(client_config, "applicationUrl", "ui.neptune.ai")

        swagger_client = MagicMock()
        swagger_client.api.getClientConfig.return_value.response.return_value.result = client_config
        swagger_client_factory.return_value = swagger_client

        return swagger_client
